from __future__ import annotations

from typing import TYPE_CHECKING

from libqtile.command.base import expose_command
from libqtile.layout.base import _SimpleLayoutBase
from libqtile.log_utils import logger

if TYPE_CHECKING:
    from typing import Any, Self

    from libqtile.backend.base import Window
    from libqtile.group import _Group

    Rect = tuple[int, int, int, int]


GOLDEN_RATIO = 1.618


class Spiral(_SimpleLayoutBase):
    """
    A mathematical layout.

    Renders windows in a spiral form by splitting the screen based on a selected ratio.
    The direction of the split is changed every time in a defined order resulting in a
    spiral formation.

    The main window can be sized with ``lazy.layout.grow_main()`` and
    ``lazy.layout.shrink_main()``. All other windows are sized by
    ``lazy.layout.increase_ratio()`` and ``lazy.layout.decrease_ratio()``.

    NB if ``main_pane_ratio`` is not set then it will also be adjusted according to ``ratio``.
    However, as soon ``shrink_main()`` or ``grow_main()`` have been called once then the
    master pane will only change size following further calls to those methods.

    Users are able to choose the location of the main (i.e. largest) pane and the direction
    of the rotation.

    Some examples:

    ``main_pane="left", clockwise=True``

    ::

        ----------------------
        |1        |2         |
        |         |          |
        |         |          |
        |         |----------|
        |         |5 |6 |3   |
        |         |-----|    |
        |         |4    |    |
        ----------------------

    ``main_pane="top", clockwise=False``

    ::

        ----------------------
        |1                   |
        |                    |
        |                    |
        |--------------------|
        |2        |5    |4   |
        |         |----------|
        |         |3         |
        ----------------------

    """

    split_ratio: float

    defaults = [
        ("border_focus", "#0000ff", "Border colour(s) for the focused window."),
        ("border_normal", "#000000", "Border colour(s) for un-focused windows."),
        ("border_width", 1, "Border width."),
        ("border_on_single", True, "Draw border when there is only one window."),
        ("margin", 0, "Margin of the layout (int or list of ints [N E S W])"),
        ("ratio", 1 / GOLDEN_RATIO, "Ratio of the tiles"),
        (
            "main_pane_ratio",
            None,
            "Ratio for biggest window or 'None' to use same ratio for all windows.",
        ),
        ("ratio_increment", 0.1, "Amount to increment per ratio increment"),
        ("main_pane", "left", "Location of biggest window 'top', 'bottom', 'left', 'right'"),
        ("clockwise", True, "Direction of spiral"),
        (
            "new_client_position",
            "top",
            "Place new windows: "
            " 'after_current' - after the active window,"
            " 'before_current' - before the active window,"
            " 'top' - in the main pane,"
            " 'bottom '- at the bottom of the stack. NB windows that are added too low in the stack"
            " may be hidden if there is no remaining space in the spiral.",
        ),
    ]

    def __init__(self, **config):
        _SimpleLayoutBase.__init__(self, **config)
        self.add_defaults(Spiral.defaults)
        self.dirty = True  # need to recalculate
        self.layout_info = []
        self.last_size = None
        self.last_screen = None
        self.initial_ratio = self.ratio
        self.initial_main_pane_ratio = self.main_pane_ratio
        self.main_pane = self.main_pane.lower()
        if self.main_pane not in ["top", "left", "bottom", "right"]:
            logger.warning(
                "Unknown main_pane location: %s. Defaulting to 'left'.", self.main_pane
            )
            self.main_pane = "left"

        # Calculate the order of transformations required based on position of main pane
        # and rotation direction
        # Lists are longer so we can pick any side and have the next 4 transformations
        if self.clockwise:
            order = ["left", "top", "right", "bottom", "left", "top", "right"]
        else:
            order = ["left", "bottom", "right", "top", "left", "bottom", "right"]

        idx = order.index(self.main_pane)
        self.splits = order[idx : idx + 4]

    def clone(self, group: _Group) -> Self:
        return _SimpleLayoutBase.clone(self, group)

    def add_client(self, client: Window) -> None:  # type: ignore[override]
        self.dirty = True
        self.clients.add_client(client, client_position=self.new_client_position)

    def remove(self, w: Window) -> Window | None:
        self.dirty = True
        return _SimpleLayoutBase.remove(self, w)

    def configure(self, win, screen):
        # force recalc
        if not self.last_screen or self.last_screen != screen:
            self.last_screen = screen
            self.dirty = True

        if self.last_size and not self.dirty:
            if screen.width != self.last_size[0] or screen.height != self.last_size[1]:
                self.dirty = True

        if self.dirty:
            self.layout_info = self.get_spiral(screen.x, screen.y, screen.width, screen.height)
            self.dirty = False

        try:
            idx = self.clients.index(win)
        except ValueError:
            win.hide()
            return

        try:
            x, y, w, h = self.layout_info[idx]
        # IndexError will arise if we're unable to create a window due to the dimensions
        # being too small. If that's the case, hide the window.
        except IndexError:
            win.hide()
            return

        if win.has_focus:
            bc = self.border_focus
        else:
            bc = self.border_normal

        (x, y, w, h), margins = self._fix_double_margins(x, y, w, h)

        if len(self.clients) == 1 and not self.border_on_single:
            border_width = 0
        else:
            border_width = self.border_width

        win.place(
            x,
            y,
            w - border_width * 2,
            h - border_width * 2,
            border_width,
            bc,
            margin=margins,
        )
        win.unhide()

    def split_left(self, rect: Rect) -> tuple[Rect, Rect]:
        rect_x, rect_y, rect_w, rect_h = rect
        win_w = int(rect_w * self.split_ratio)
        win_h = rect_h
        win_x = rect_x
        win_y = rect_y
        rect_x = win_x + win_w
        rect_y = win_y
        rect_w = rect_w - win_w
        return (win_x, win_y, win_w, win_h), (rect_x, rect_y, rect_w, rect_h)

    def split_right(self, rect: Rect) -> tuple[Rect, Rect]:
        rect_x, rect_y, rect_w, rect_h = rect
        win_w = int(rect_w * self.split_ratio)
        win_h = rect_h
        win_x = rect_x + (rect_w - win_w)
        win_y = rect_y
        rect_x = win_x - (rect_w - win_w)
        rect_y = win_y
        rect_w = rect_w - win_w
        return (win_x, win_y, win_w, win_h), (rect_x, rect_y, rect_w, rect_h)

    def split_top(self, rect: Rect) -> tuple[Rect, Rect]:
        rect_x, rect_y, rect_w, rect_h = rect
        win_w = rect_w
        win_h = int(rect_h * self.split_ratio)
        win_x = rect_x
        win_y = rect_y
        rect_x = win_x
        rect_y = win_y + win_h
        rect_h = rect_h - win_h
        return (win_x, win_y, win_w, win_h), (rect_x, rect_y, rect_w, rect_h)

    def split_bottom(self, rect: Rect) -> tuple[Rect, Rect]:
        rect_x, rect_y, rect_w, rect_h = rect
        win_w = rect_w
        win_h = int(rect_h * self.split_ratio)
        win_x = rect_x
        win_y = rect_y + (rect_h - win_h)
        rect_x = win_x
        rect_y = win_y - (rect_h - win_h)
        rect_h = rect_h - win_h
        return (win_x, win_y, win_w, win_h), (rect_x, rect_y, rect_w, rect_h)

    def _fix_double_margins(
        self, win_x: int, win_y: int, win_w: int, win_h: int
    ) -> tuple[Rect, list[int]]:
        """Prevent doubling up of margins by halving margins for internal margins."""
        if isinstance(self.margin, int):
            margins = [self.margin] * 4
        else:
            margins = self.margin

        # Top
        if win_y - margins[0] > self.last_screen.y:
            win_y -= margins[0] // 2
            win_h += margins[0] // 2

        # Right
        if win_x + win_w + margins[1] < (self.last_screen.x + self.last_screen.width):
            win_w += margins[1] // 2

        # Bottom
        if win_y + win_h + margins[2] < (self.last_screen.y + self.last_screen.height):
            win_h += margins[2] // 2

        # Left
        if win_x - margins[3] > self.last_screen.x:
            win_x -= margins[3] // 2
            win_w += margins[3] // 2

        return (win_x, win_y, win_w, win_h), margins

    def has_invalid_size(self, win: Rect) -> bool:
        """
        Checks if window would have an invalid size.

        A window that would have negative height or width (after adjusting for margins and borders)
        will return True.
        """
        if isinstance(self.margin, int):
            margin = [self.margin] * 4
        else:
            margin = self.margin

        return any(
            [
                win[2] <= margin[1] + margin[3] + 2 * self.border_width,
                win[3] <= margin[0] + margin[2] + 2 * self.border_width,
            ]
        )

    def get_spiral(self, x, y, width, height) -> list[Rect]:
        """
        Calculates positions of windows in the spiral.

        Returns a list of tuples (x, y, w, h) for positioning windows.
        """
        num_windows = len(self.clients)
        direction = 0
        spiral = []
        rect = (x, y, width, height)

        for c in range(num_windows):
            if c == 0 and self.main_pane_ratio is not None:
                self.split_ratio = self.main_pane_ratio
            else:
                self.split_ratio = self.ratio

            # If there's another window to draw after this one then we need to
            # split the current rect, if not, the window can take the full rect
            split = c < (num_windows - 1)
            if not split:
                spiral.append(rect)
                continue

            # Get the dimensions of the window and remaining rect
            # Calls self.split_[direction name]
            win, new_rect = getattr(self, f"split_{self.splits[direction]}")(rect)

            # If the window would have negative/zero dimensions then it can't be displayed
            if self.has_invalid_size(win):
                # Use the available rect from before the split
                spiral.append(rect)
                break

            spiral.append(win)

            direction = (direction + 1) % 4
            rect = new_rect

        return spiral

    def info(self) -> dict[str, Any]:
        d = _SimpleLayoutBase.info(self)
        focused = self.clients.current_client
        d["ratio"] = self.ratio
        d["focused"] = focused.name if focused else None
        d["layout_info"] = self.layout_info
        d["main_pane"] = self.main_pane
        d["clockwise"] = self.clockwise
        return d

    @expose_command("up")
    def previous(self) -> None:
        _SimpleLayoutBase.previous(self)

    @expose_command("down")
    def next(self) -> None:
        _SimpleLayoutBase.next(self)

    def _set_ratio(self, prop: str, value: float):
        if not (0 <= value <= 1):
            logger.warning(
                "Invalid value for %s: %s. Value must be between 0 and 1.", prop, value
            )
            return

        setattr(self, prop, value)
        # Force layout to be recalculated
        self.dirty = True
        self.group.layout_all()

    @expose_command()
    def shuffle_down(self):
        if self.clients:
            self.clients.shuffle_down()
            self.group.layout_all()

    @expose_command()
    def shuffle_up(self):
        if self.clients:
            self.clients.shuffle_up()
            self.group.layout_all()

    @expose_command()
    def decrease_ratio(self):
        """Decrease spiral ratio."""
        self._set_ratio("ratio", self.ratio - self.ratio_increment)

    @expose_command()
    def increase_ratio(self):
        """Increase spiral ratio."""
        self._set_ratio("ratio", self.ratio + self.ratio_increment)

    @expose_command()
    def shrink_main(self):
        """Shrink the main window."""
        if self.main_pane_ratio is None:
            self.main_pane_ratio = self.ratio

        self._set_ratio("main_pane_ratio", self.main_pane_ratio - self.ratio_increment)

    @expose_command()
    def grow_main(self):
        """Grow the main window."""
        if self.main_pane_ratio is None:
            self.main_pane_ratio = self.ratio

        self._set_ratio("main_pane_ratio", self.main_pane_ratio + self.ratio_increment)

    @expose_command()
    def set_ratio(self, ratio: float):
        """Set the ratio for all windows."""
        self._set_ratio("ratio", ratio)

    @expose_command()
    def set_master_ratio(self, ratio: float):
        """Set the ratio for the main window."""
        self._set_ratio("main_pane_ratio", ratio)

    @expose_command()
    def reset(self):
        """Reset ratios to values set in config."""
        self.ratio = self.initial_ratio
        self.main_pane_ratio = self.initial_main_pane_ratio
        # Force layout to be recalculated
        self.dirty = True
        self.group.layout_all()



from __future__ import annotations

import codeop
import re
import socket
import sys
from time import sleep
from typing import TYPE_CHECKING

try:
    from prompt_toolkit import PromptSession
    from prompt_toolkit.completion import Completer, Completion
    from prompt_toolkit.history import InMemoryHistory
    from prompt_toolkit.key_binding import KeyBindings
    from prompt_toolkit.patch_stdout import patch_stdout

    HAS_PT = True
except (ImportError, ModuleNotFoundError):
    HAS_PT = False

from libqtile.interactive.repl import COMPLETION_REQUEST, REPL_PORT, TERMINATOR
from libqtile.scripts.cmd_obj import cmd_obj

if TYPE_CHECKING:
    pass

HOST = "localhost"


class Command:
    """Wrapper to call commands via command interface."""

    def __init__(self, command, obj_spec=["root"], *args, **kwargs):
        self.function = command
        self.socket = None
        self.args = args
        self.kwargs = kwargs
        self.obj_spec = obj_spec
        self.info = False

    def __call__(self):
        return cmd_obj(self)


# Calls to start and stop the qtile REPL server
# Sends commands via qtile cmd-obj
start_server = Command("start_repl_server")
stop_server = Command("stop_repl_server")


def is_code_complete(text: str) -> bool:
    """Method to verify that we have a valid code block."""
    try:
        code_obj = codeop.compile_command(text, symbol="exec")

        # Incomplete (e.g. after 'def foo():')
        if code_obj is None:
            return False

        # Only treat as complete if there's a double blank line at the end for compound blocks
        lines = text.rstrip("\n").splitlines()
        return len(lines) <= 1 or text.endswith("\n\n")
    except (SyntaxError, OverflowError, ValueError, TypeError):
        return True


def read_full_response(sock, end_marker=f"{TERMINATOR}\n"):
    """Function to read data from socket until termination marker."""
    buffer = ""
    while True:
        data = sock.recv(4096).decode()

        if not data:
            # connection closed without end marker
            break

        buffer += data
        if end_marker in buffer:
            # Split off the marker and return only the response text
            response, _, _ = buffer.partition(end_marker)
            return response

    # connection closed without end marker
    return buffer


def start_repl(_args):
    if not HAS_PT:
        sys.exit("You need to install prompt_toolkit to use the REPL client.")

    # Start the repl server in qtile and find port number
    start_server()

    # We need to wait until server is up an running before continuing
    retry_count = 0
    while retry_count < 5:
        try:
            sock = socket.create_connection((HOST, REPL_PORT))
            break
        except ConnectionRefusedError:
            retry_count += 1
            sleep(0.5)

    if retry_count == 5:
        print("Unable to connect to REPL server. Exiting...")
        stop_server()
        return

    try:
        # Create the objects needed for the client
        class SocketCompleter(Completer):
            def __init__(self, sock):
                self.sock = sock

            def get_completions(self, document, _complete_event):
                text_before_cursor = document.text_before_cursor

                # Extract the current word or attribute expression (e.g., qtile.cur)
                match = re.search(r"([\w\.]+)$", text_before_cursor)
                if not match:
                    return

                word = match.group(1)
                start_position = -len(word)

                # Send only the word to the REPL server
                self.sock.sendall(f"{COMPLETION_REQUEST}{word}\n{TERMINATOR}\n".encode())

                # Read completions from server and filter out empty strings
                data = read_full_response(self.sock)
                options = list(filter(None, data.strip().split(",")))

                # No suggestions so return early
                if not options:
                    return

                for opt in options:
                    # opt is the full completion (e.g. 'qtile.current_layout')
                    yield Completion(opt, start_position=start_position)

        kb = KeyBindings()
        completer = SocketCompleter(sock)

        # Create a session instance
        session = PromptSession(
            completer=completer,
            key_bindings=kb,
            complete_while_typing=False,
            multiline=True,
            prompt_continuation=lambda *args, **kwargs: "... ",
            history=InMemoryHistory(),
        )

        # Create a handler for the Enter key so we can deal with multiline input
        @kb.add("enter")
        def _(event):
            buffer = event.app.current_buffer
            text = buffer.document.text

            if is_code_complete(text):
                # Save input line to print after
                full_block = f"{text}\n{TERMINATOR}\n"

                # Submit to server
                sock.sendall(full_block.encode())

                # Save our code to the history as `buffer.reset()`
                # would otherwise prevent that from happening
                session.history.append_string(text)

                # Clear buffer before reading response
                buffer.reset()

                # Read and print server response
                response = read_full_response(sock)

                # Echo input and response manually
                text = text.replace("\n", "\n... ")
                print(f">>> {text}")
                if response.strip():
                    print(response, end="\n", flush=True)
            else:
                buffer.insert_text("\n")  # Insert a newline instead

        with patch_stdout():
            # Read the welcome message from the server.
            print(read_full_response(sock), end="", flush=True)

            while True:
                try:
                    session.prompt(">>> ")
                except KeyboardInterrupt:
                    print("\nExiting.")
                    break

    finally:
        sock.close()
        stop_server()


def add_subcommand(subparsers, parents):
    parser = subparsers.add_parser("repl", parents=parents, help="Run a qtile REPL session.")
    parser.set_defaults(func=start_repl)



"""
The objects in the command graph and command resolution on the objects
"""

from __future__ import annotations

import abc
import asyncio
import inspect
import sys
import traceback
from typing import TYPE_CHECKING

from libqtile.log_utils import logger
from libqtile.utils import create_task

if TYPE_CHECKING:
    from collections.abc import Callable

    from libqtile.command.graph import SelectorType

    ItemT = tuple[bool, list[str | int]] | None


def allow_when_locked(func: Callable) -> Callable:
    """
    Decorator to expose methods which can be run when the
    session is locked.
    """
    setattr(func, "_allow_when_locked", True)
    return func


def expose_command(name: Callable | str | list[str] | None = None) -> Callable:
    """
    Decorator to expose methods to the command interface.

    The exposed command will have the name of the defined method.

    Methods can also be exposed via multiple names by passing the names to this
    decorator.

    e.g. if a layout wants "up" and "previous" to call the
    same method:

    @expose_command("previous")
    def up(self):
        ...

    `up` will be exposed as `up` and `previous`.

    Multiple names can be passed as a list.
    """

    def wrapper(func: Callable):
        setattr(func, "_cmd", True)
        if name is not None:
            if not hasattr(func, "_mapping"):
                setattr(func, "_mapping", list())
            if isinstance(name, list):
                func._mapping += name  # type:ignore
            elif isinstance(name, str):
                func._mapping.append(name)  # type:ignore
            else:
                logger.error("Unexpected value received in command decorator: %s", name)
        return func

    # If the decorator is added with no parentheses then we should treat it
    # as if it had been i.e. expose the decorated method
    if callable(name):
        func = name
        name = None
        return wrapper(func)

    return wrapper


class SelectError(Exception):
    """Error raised in resolving a command graph object"""

    def __init__(self, err_string: str, name: str, selectors: list[SelectorType]):
        super().__init__(f"{err_string}, name: {name}, selectors: {selectors}")
        self.name = name
        self.selectors = selectors


class CommandError(Exception):
    """Error raised in resolving a command"""


class CommandException(Exception):
    """Error raised while executing a command"""


class CommandObject(metaclass=abc.ABCMeta):
    """Base class for objects that expose commands

    Any command to be exposed should be decorated with
    `@expose_command()` (classes that are not explicitly
    inheriting from CommandObject will need to import the module)
    A CommandObject should also implement `._items()` and `._select()` methods
    (c.f. docstring for `.items()` and `.select()`).
    """

    def __new__(cls, *args, **kwargs):
        # Check which level of command object has been parsed
        # This test ensures inherited classes don't stop additional
        # methods from being exposed.
        # For example, if widget.TextBox has already been parsed, a subsequent
        # call to initialise a new TextBox will return here. However, if a user
        # subclasses TextBox for a new widget then that new widget will still
        # be parsed here to check for new commands.
        if getattr(cls, "_command_object", "") == cls.__name__:
            super().__new__(cls)

        commands = {}
        cmd_s = set()

        # We need to iterate over the class's inherited classes in reverse order
        # We reverse the order so the exposed command will always be the latest
        # definition of the method.
        for c in reversed(list(cls.__mro__)):
            for method_name in list(c.__dict__.keys()):
                method = getattr(c, method_name, None)

                if method is None:
                    continue

                # If the command has been exposed, add it to our dictionary
                # If the method name is already in our dictionary then bind the
                # latest definition to that command
                if hasattr(method, "_cmd") or method_name in commands:
                    commands[method_name] = method

                # Expose additional names
                for mapping in getattr(method, "_mapping", list()):
                    setattr(cls, mapping, method)
                    commands[mapping] = method

        if cmd_s:
            names = ", ".join(cmd_s)
            msg = (
                f"The use of the 'cmd_' prefix to expose commands via IPC "
                f"is deprecated. Methods should use the "
                f"@expose_command() decorator instead. "
                f"Please update: {names}"
            )
            logger.warning("Deprecation Warning: %s", msg)

        # Record the object as being parsed.
        cls._command_object = cls.__name__

        # Store list of exposed commands
        cls._commands = commands

        return super().__new__(cls)

    def select(self, selectors: list[SelectorType]) -> CommandObject:
        """Return a selected object

        Recursively finds an object specified by a list of `(name, selector)`
        items.

        Raises SelectError if the object does not exist.
        """
        obj: CommandObject = self
        for name, selector in selectors:
            root, items = obj.items(name)
            # if non-root object and no selector given
            if root is False and selector is None:
                raise SelectError("", name, selectors)
            # if no items in container, but selector is given
            if items is None and selector is not None:
                raise SelectError("", name, selectors)
            # if selector is not in the list of contained items
            if items is not None and selector and selector not in items:
                raise SelectError("", name, selectors)

            maybe_obj = obj._select(name, selector)
            if maybe_obj is None:
                raise SelectError("", name, selectors)
            obj = maybe_obj
        return obj

    @expose_command()
    def items(self, name: str) -> tuple[bool, list[str | int] | None]:
        """
        Build a list of contained items for the given item class.

        Exposing this allows __qsh__ to navigate the command graph.

        Returns a tuple `(root, items)` for the specified item class, where:

            root: True if this class accepts a "naked" specification without an
            item seletion (e.g. "layout" defaults to current layout), and False
            if it does not (e.g. no default "widget").

            items: a list of contained items
        """
        ret = self._items(name)
        if ret is None:
            # Not finding information for a particular item class is OK here;
            # we don't expect layouts to have a window, etc.
            return False, None
        return ret

    @abc.abstractmethod
    def _items(self, name) -> ItemT:
        """Generate the items for a given

        Same return as `.items()`. Return `None` if name is not a valid item
        class.
        """

    @abc.abstractmethod
    def _select(self, name: str, sel: str | int | None) -> CommandObject | None:
        """Select the given item of the given item class

        This method is called with the following guarantees:
            - `name` is a valid selector class for this item
            - `sel` is a valid selector for this item
            - the `(name, sel)` tuple is not an "impossible" combination (e.g. a
              selector is specified when `name` is not a containment object).

        Return None if no such object exists
        """

    def command(self, name: str) -> Callable | None:
        """Return the command with the given name

        Parameters
        ----------
        name: str
            The name of the command to fetch.

        """
        return self._commands.get(name)

    @expose_command()
    def commands(self) -> list[str]:
        """
        Returns a list of possible commands for this object

        Used by __qsh__ for command completion and online help
        """
        return sorted([cmd for cmd in self._commands])

    @expose_command()
    def doc(self, name) -> str:
        """Returns the documentation for a specified command name

        Used by __qsh__ to provide online help.
        """
        if name in self.commands():
            command = self.command(name)
            assert command
            signature = self._get_command_signature(command)
            spec = name + signature
            htext = inspect.getdoc(command) or ""
            return spec + "\n" + htext
        raise CommandError(f"No such command: {name}")

    def _get_command_signature(self, command: Callable) -> str:
        signature = inspect.signature(command)
        args = list(signature.parameters)
        if args and args[0] == "self":
            args = args[1:]
            parameters = [signature.parameters[arg] for arg in args]
            signature = signature.replace(parameters=parameters)
        return str(signature)

    @expose_command()
    def eval(self, code: str) -> str | None:
        """Evaluates code in the same context as this function

        Return a string representing the return value of eval, or None if
        exec was used instead.
        """
        globals_ = vars(sys.modules[self.__module__])
        try:
            return str(eval(code, globals_, locals()))
        except SyntaxError:
            exec(code, globals_, locals())
            return None

    @expose_command()
    def function(self, function, *args, **kwargs) -> asyncio.Task | None:
        """Call a function with current object as argument"""
        try:
            if inspect.iscoroutinefunction(function):
                return create_task(function(self, *args, **kwargs))
            else:
                return function(self, *args, **kwargs)
        except Exception:
            error = traceback.format_exc()
            logger.error('Exception calling "%s":\n%s', function, error)
            return None



from __future__ import annotations

from enum import IntEnum, auto
from typing import TYPE_CHECKING

from libqtile import hook
from libqtile.log_utils import logger

if TYPE_CHECKING:
    from collections.abc import Callable
    from typing import Any

    from libqtile.backend.base.core import Core
    from libqtile.backend.base.window import Window
    from libqtile.core.manager import Qtile


class InhibitorType(IntEnum):
    GLOBAL = auto()
    OPEN = auto()
    APPLICATION = auto()
    VISIBLE = auto()
    FOCUS = auto()
    FULLSCREEN = auto()
    UNSET = auto()


inhibitor_map = {
    "open": InhibitorType.OPEN,
    "visible": InhibitorType.VISIBLE,
    "focus": InhibitorType.FOCUS,
    "fullscreen": InhibitorType.FULLSCREEN,
}


class Inhibitor:
    def __init__(
        self,
        qtile: Qtile,
        window: Window | None = None,
        inhibitor_type: InhibitorType = InhibitorType.UNSET,
    ):
        if (
            inhibitor_type not in (InhibitorType.GLOBAL, InhibitorType.APPLICATION)
            and window is None
        ):
            raise ValueError("Inhibitor created with invalid arguments.")

        self.qtile = qtile
        self.window = window
        self.inhibitor_type = inhibitor_type

    def check(self) -> bool:
        active = False

        match self.inhibitor_type:
            # Global inhibitor is always active
            case InhibitorType.GLOBAL:
                active = True
            # Inhibitor is removed when client is killed so if it exists in the list then the client
            # must be open
            case InhibitorType.OPEN:
                active = True
            case InhibitorType.FOCUS:
                active = bool(
                    self.qtile.current_window and self.qtile.current_window == self.window
                )
            case InhibitorType.FULLSCREEN:
                active = bool(self.window and self.window.fullscreen)
            case InhibitorType.VISIBLE:
                active = bool(self.window and self.window.is_visible())

        return active

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Inhibitor):
            raise NotImplementedError

        return (self.window, self.inhibitor_type) == (other.window, other.inhibitor_type)

    def __repr__(self) -> str:
        name = self.window.name if self.window else "no window"
        mode = self.inhibitor_type.name.lower()
        active = "ACTIVE" if self.check() else "inactive"
        return f"<window={name} type={mode} status={active}>"


class IdleInhibitorManager[TInhibitor: Inhibitor]:
    def __init__(self, core: Core) -> None:
        self.core = core
        self.inhibitors: list[TInhibitor] = []

    def _convert_inhibitor(self, inhibitor: Inhibitor) -> TInhibitor:
        """
        Convert a base Inhibitor into the manager's concrete inhibitor type.
        Base implementation only works if TInhibitor *is* Inhibitor.
        """
        return inhibitor  # type: ignore[return-value]

    def set_hooks(self) -> None:
        hook.subscribe.focus_change(self.check)
        hook.subscribe.startup(self.update_user_inhibitors)

    def sort(self) -> None:
        self.inhibitors.sort(key=lambda x: x.inhibitor_type)

    def add_inhibitor_safe(self, inhibitor: TInhibitor) -> bool:
        if inhibitor not in self.inhibitors:
            self.inhibitors.append(inhibitor)
            self.sort()
            self.check()
            return True

        return False

    def remove_inhibitor(
        self, remove: Callable[[TInhibitor], bool], skip_check: bool = False
    ) -> bool:
        old = len(self.inhibitors)
        self.inhibitors = [o for o in self.inhibitors if not remove(o)]
        removed = len(self.inhibitors) < old
        if removed and not skip_check:
            self.check()

        return removed

    def add_window_inhibitor(self, window: Window, inhibitor_type: str) -> None:
        itype = inhibitor_map.get(inhibitor_type)
        if itype is None:
            logger.error("Unexpected inhibitor type {inhibitor_type}.")
            return

        inhibitor = Inhibitor(qtile=self.core.qtile, window=window, inhibitor_type=itype)
        self.add_inhibitor_safe(self._convert_inhibitor(inhibitor))

    def remove_window_inhibitor(self, window: Window) -> None:
        self.remove_inhibitor(lambda o: o.window == window)

    def add_global_inhibitor(self) -> None:
        inhibitor = Inhibitor(qtile=self.core.qtile, inhibitor_type=InhibitorType.GLOBAL)
        self.add_inhibitor_safe(self._convert_inhibitor(inhibitor))

    def remove_global_inhibitor(self) -> None:
        self.remove_inhibitor(lambda o: o.inhibitor_type == InhibitorType.GLOBAL)

    def check(self) -> None:
        inhibited = any(o.check() for o in self.inhibitors)
        self.core.inhibited = inhibited

    def update_user_inhibitors(self) -> None:
        self.remove_inhibitor(
            lambda o: o.inhibitor_type != InhibitorType.APPLICATION, skip_check=True
        )
        for win in self.core.qtile.windows_map.values():
            if isinstance(win, Window):
                win.add_config_inhibitors()



import locale

from libqtile.confreader import ConfigError
from libqtile.log_utils import logger
from libqtile.widget.gen_poll_url import GenPollUrl

_DEFAULT_CURRENCY = str(locale.localeconv()["int_curr_symbol"])
_DEFAULT_SYMBOL = str(locale.localeconv()["currency_symbol"])


class CryptoTicker(GenPollUrl):
    """
    A cryptocurrency ticker widget, data provided by the coinbase.com or the binance.com
    API. Defaults to displaying currency in whatever the current locale is. Examples:

        # display the average price of bitcoin in local currency
        widget.CryptoTicker()

        # display it in Euros:
        widget.CryptoTicker(currency="EUR")

        # or a different cryptocurrency!
        widget.CryptoTicker(crypto="ETH")

        # change the currency symbol:
        widget.CryptoTicker(currency="EUR", symbol="€")

        # display from Binance API
        widget.CryptoTicker(api="binance", currency="USDT")

    Widget requirements: aiohttp_.

    .. _aiohttp: https://pypi.org/project/aiohttp/
    """

    QUERY_URL_DICT = {
        "coinbase": (
            "https://api.coinbase.com/v2/prices/{}-{}/spot",
            lambda x: float(x["data"]["amount"]),
        ),
        "binance": (
            "https://api.binance.com/api/v3/ticker/price?symbol={}{}",
            lambda x: float(x["price"]),
        ),
    }

    defaults = [
        (
            "currency",
            _DEFAULT_CURRENCY.strip(),
            "The baseline currency that the value of the crypto is displayed in.",
        ),
        ("symbol", _DEFAULT_SYMBOL, "The symbol for the baseline currency."),
        ("crypto", "BTC", "The cryptocurrency to display."),
        ("format", "{crypto}: {symbol}{amount:.2f}", "Display string formatting."),
        ("api", "coinbase", "API that provides the data."),
    ]

    def __init__(self, **config):
        GenPollUrl.__init__(self, **config)
        self.add_defaults(CryptoTicker.defaults)

        # set up USD as the currency if no locale is set
        if self.currency == "":
            self.currency = "USD"
        # set up $ as the symbol if no locale is set
        if self.symbol == "":
            self.symbol = "$"

    def _configure(self, qtile, bar):
        try:
            GenPollUrl._configure(self, qtile, bar)
            self.query_url = self.QUERY_URL_DICT[self.api][0]
        except KeyError:
            apis = sorted(self.QUERY_URL_DICT.keys())
            logger.error(
                "%s is not a valid API. Use one of the list: %s.",
                self.api,
                apis,
            )
            raise ConfigError("Unknown provider passed as 'api' to CryptoTicker")

    @property
    def url(self):
        return self.query_url.format(self.crypto, self.currency)

    def parse(self, body):
        variables = dict()
        variables["crypto"] = self.crypto
        variables["symbol"] = self.symbol
        variables["amount"] = self.QUERY_URL_DICT[self.api][1](body)

        return self.format.format(**variables)



import shutil
import subprocess
import sys
import tempfile
import traceback
from os import environ, path

from libqtile import confreader
from libqtile.utils import get_config_file
from libqtile.widget.import_error import ImportErrorWidget


class CheckError(Exception):
    pass


def type_check_config_vars(tempdir, config_name):
    # write a .pyi file to tempdir:
    f = open(path.join(tempdir, config_name + ".pyi"), "w")
    f.write(confreader.config_pyi_header)
    for name, type_ in confreader.Config.__annotations__.items():
        f.write(name)
        f.write(": ")
        f.write(type_)
        f.write("\n")
    f.close()

    # need to tell python to look in pwd for modules
    newenv = environ.copy()
    newenv["PYTHONPATH"] = newenv.get("PYTHONPATH", "") + ":"

    p = subprocess.Popen(
        ["stubtest", "--concise", config_name],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        cwd=tempdir,
        text=True,
        env=newenv,
    )
    stdout, stderr = p.communicate()
    missing_vars = []
    for line in (stdout + stderr).split("\n"):
        # filter out stuff that users didn't specify; they'll be imported from
        # the default config
        if "is not present at runtime" in line:
            missing_vars.append(line.split()[0])

    # write missing vars to a tempfile
    whitelist = open(path.join(tempdir, "stubtest_whitelist"), "w")
    for var in missing_vars:
        whitelist.write(var)
        whitelist.write("\n")
    whitelist.close()

    p = subprocess.Popen(
        [
            "stubtest",
            # ignore variables that the user creates in their config that
            # aren't in our default config list
            "--ignore-missing-stub",
            # use our whitelist to ignore stuff users didn't specify
            "--whitelist",
            whitelist.name,
            config_name,
        ],
        cwd=tempdir,
        text=True,
        env=newenv,
    )
    p.wait()
    if p.returncode != 0:
        raise CheckError()


def type_check_config_args(config_file):
    newenv = environ.copy()
    try:
        subprocess.check_call(["mypy", config_file], env=newenv)
        print("Config file type checking succeeded!")
    except subprocess.CalledProcessError as e:
        print(f"Config file type checking failed: {e}")
        raise CheckError()


def check_deps() -> None:
    ok = True

    for dep in ["mypy", "stubtest"]:
        if shutil.which(dep) is None:
            print(f"{dep} was not found in PATH. Please install it, add to PATH and try again.")
            ok = False

    if not ok:
        raise CheckError()


def check_config(args):
    args.configfile = path.abspath(args.configfile)
    print(f"Checking Qtile config at: {args.configfile}")
    print("Checking if config is valid python...")

    try:
        # can we load the config?
        config = confreader.Config(args.configfile)
        config.load()
        config.validate()
    except Exception:
        print(traceback.format_exc())
        sys.exit("Errors found in config. Exiting check.")

    print("Checking widget dependencies...")
    widgets = []
    for screen in config.screens:
        for bar_name in ("top", "bottom", "left", "right"):
            bar = getattr(screen, bar_name, None)
            if bar and hasattr(bar, "widgets"):
                widgets.extend(bar.widgets)

    errors = {
        w.widget_class: w.missing_dependencies
        for w in widgets
        if isinstance(w, ImportErrorWidget)
    }

    if errors:
        print("The following widgets have import errors:")
    for widget, deps in errors.items():
        print(f" {widget}: {', '.join(deps)}")

    try:
        check_deps()
    except CheckError:
        print("Missing dependencies. Skipping type checking.")
    else:
        # need to do all the checking in a tempdir because we need to write stuff
        # for stubtest
        print("Type checking config file...")
        valid = True
        with tempfile.TemporaryDirectory() as tempdir:
            shutil.copytree(
                path.dirname(args.configfile),
                tempdir,
                dirs_exist_ok=True,
                symlinks=True,
                ignore=shutil.ignore_patterns(
                    ".git",
                    ".venv",
                    ".venv*",
                    "__pycache__",
                    ".mypy_cache",
                    ".ruff_cache",
                    "test",
                    "docs",
                    "nix",
                    "stubs",
                ),
            )
            tmp_path = path.join(tempdir, path.basename(args.configfile))

            # are the top level config variables the right type?
            module_name = path.splitext(path.basename(args.configfile))[0]
            try:
                type_check_config_vars(tempdir, module_name)
            except CheckError:
                valid = False

            # are arguments passed to qtile APIs correct?
            try:
                type_check_config_args(tmp_path)
            except CheckError:
                valid = False

        if valid:
            print("Your config can be loaded by Qtile.")
        else:
            sys.exit(
                "Your config is valid python but has type checking errors. This may result in unexpected behaviour."
            )


def add_subcommand(subparsers, parents):
    parser = subparsers.add_parser(
        "check", parents=parents, help="Check a configuration file for errors."
    )
    parser.add_argument(
        "-c",
        "--config",
        action="store",
        default=get_config_file(),
        dest="configfile",
        help="Use the specified configuration file.",
    )
    parser.set_defaults(func=check_config)



"""
Command-line top like for qtile
"""

import curses
import linecache
import os
import sys
import time

from libqtile import ipc
from libqtile.command import client, interface

# These imports are here because they are not supported in pypy.
# having them at the top of the file causes problems when running any
# of the other scripts.
try:
    import tracemalloc
    from tracemalloc import Snapshot

    ENABLED = True
except ModuleNotFoundError:
    ENABLED = False


class TraceCantStart(Exception):
    pass


def get_trace(c):
    (started, path) = c.tracemalloc_dump()
    if not started:
        c.tracemalloc_toggle()
        (started, path) = c.tracemalloc_dump()
        if not started:
            raise TraceCantStart

    return Snapshot.load(path)


def filter_snapshot(snapshot):
    return snapshot.filter_traces(
        (
            tracemalloc.Filter(False, "<frozen importlib._bootstrap>"),
            tracemalloc.Filter(False, "<unknown>"),
        )
    )


def get_stats(scr, c, group_by="lineno", limit=10, seconds=1.5):
    (max_y, max_x) = scr.getmaxyx()
    curses.init_pair(1, curses.COLOR_GREEN, curses.COLOR_BLACK)
    while True:
        scr.addstr(0, 0, f"Qtile - Top {limit} lines")
        scr.addstr(
            1,
            0,
            "{:<3s} {:<40s} {:<30s} {:<16s}".format("#", "Line", "Memory", " " * (max_x - 71)),
            curses.A_BOLD | curses.A_REVERSE,
        )

        snapshot = get_trace(c)
        snapshot = filter_snapshot(snapshot)
        top_stats = snapshot.statistics(group_by)
        cnt = 1
        for index, stat in enumerate(top_stats[:limit], 1):
            frame = stat.traceback[0]
            # replace "/path/to/module/file.py" with "module/file.py"
            filename = os.sep.join(frame.filename.split(os.sep)[-2:])
            code = ""
            line = linecache.getline(frame.filename, frame.lineno).strip()
            if line:
                code = line
            mem = f"{stat.size / 1024.0:.1f} KiB"
            filename = f"{filename}:{frame.lineno}"
            scr.addstr(cnt + 1, 0, f"{index:<3} {filename:<40} {mem:<30}")
            scr.addstr(cnt + 2, 4, code, curses.color_pair(1))
            cnt += 2

        other = top_stats[limit:]
        cnt += 2
        if other:
            size = sum(stat.size for stat in other)
            other_size = f"{len(other):d} other: {size / 1024.0:.1f} KiB"
            scr.addstr(cnt, 0, other_size, curses.A_BOLD)
            cnt += 1

        total = sum(stat.size for stat in top_stats)
        total_size = f"Total allocated size: {total / 1024.0:.1f} KiB"
        scr.addstr(cnt, 0, total_size, curses.A_BOLD)

        scr.move(max_y - 2, max_y - 2)
        scr.refresh()
        time.sleep(seconds)
        scr.erase()


def raw_stats(c, group_by="lineno", limit=10):
    snapshot = get_trace(c)
    snapshot = filter_snapshot(snapshot)
    top_stats = snapshot.statistics(group_by)

    print(f"Qtile - Top {limit} lines")
    for index, stat in enumerate(top_stats[:limit], 1):
        frame = stat.traceback[0]
        # replace "/path/to/module/file.py" with "module/file.py"
        filename = os.sep.join(frame.filename.split(os.sep)[-2:])
        print(f"#{index}: {filename}:{frame.lineno}: {stat.size / 1024.0:.1f} KiB")
        line = linecache.getline(frame.filename, frame.lineno).strip()
        if line:
            print(f"    {line}")

    other = top_stats[limit:]
    if other:
        size = sum(stat.size for stat in other)
        print(f"{len(other):d} other: {size / 1024.0:.1f} KiB")
    total = sum(stat.size for stat in top_stats)
    print(f"Total allocated size: {total / 1024.0:.1f} KiB")


def top(opts):
    if not ENABLED:
        raise Exception("Could not import tracemalloc")
    lines = opts.lines
    seconds = opts.seconds
    if opts.socket is None:
        socket = ipc.find_sockfile()
    else:
        socket = opts.socket
    c = client.InteractiveCommandClient(
        interface.IPCCommandInterface(
            ipc.Client(socket),
        ),
    )

    try:
        if not opts.raw:
            curses.wrapper(get_stats, c, limit=lines, seconds=seconds)
        else:
            raw_stats(c, limit=lines)
    except TraceCantStart:
        print("Can't start tracemalloc on qtile, check the logs")
    except KeyboardInterrupt:
        sys.exit(1)
    except curses.error:
        print("Terminal too small for curses interface.")
        raw_stats(c, limit=lines)


def add_subcommand(subparsers, parents):
    parser = subparsers.add_parser(
        "top", parents=parents, help="A top-like resource usage monitor."
    )
    parser.add_argument(
        "-L", "--lines", type=int, dest="lines", default=10, help="Number of lines to show."
    )
    parser.add_argument(
        "-r",
        "--raw",
        dest="raw",
        action="store_true",
        default=False,
        help="Output raw without curses.",
    )
    parser.add_argument(
        "-t",
        "--time",
        type=float,
        dest="seconds",
        default=1.5,
        help="Refresh rate.",
    )
    parser.add_argument(
        "-s", "--socket", type=str, dest="socket", help="Use specified socket for IPC."
    )
    parser.set_defaults(func=top)



from __future__ import annotations

import asyncio
import faulthandler
import io
import logging
import os
import pickle
import shlex
import shutil
import signal
import socket
import tempfile
import time
from collections import defaultdict
from logging.handlers import RotatingFileHandler
from pathlib import Path
from typing import TYPE_CHECKING

import libqtile
from libqtile import bar, hook, ipc, utils
from libqtile.backend import base
from libqtile.command import interface
from libqtile.command.base import CommandError, CommandException, CommandObject, expose_command
from libqtile.command.client import InteractiveCommandClient
from libqtile.command.interface import IPCCommandServer, QtileCommandInterface
from libqtile.config import (
    Click,
    Drag,
    Key,
    KeyChord,
    Match,
    Mouse,
    Output,
    Rule,
    Screen,
    ScreenRect,
)
from libqtile.config import ScratchPad as ScratchPadConfig
from libqtile.core.lifecycle import lifecycle
from libqtile.core.loop import LoopContext
from libqtile.core.state import QtileState
from libqtile.dgroups import DGroups
from libqtile.extension.base import _Extension
from libqtile.group import _Group
from libqtile.interactive.repl import repl_server
from libqtile.log_utils import logger
from libqtile.resources.sleep import inhibitor
from libqtile.scratchpad import ScratchPad
from libqtile.scripts.main import VERSION
from libqtile.utils import (
    create_task,
    get_cache_dir,
    lget,
    remove_dbus_rules,
    send_notification,
)
from libqtile.widget.base import _Widget

if TYPE_CHECKING:
    from collections.abc import Callable, Sequence
    from os import PathLike
    from typing import Any, Literal

    from libqtile.command.base import ItemT
    from libqtile.confreader import Config
    from libqtile.layout.base import Layout
    from libqtile.utils import ColorType


class Qtile(CommandObject):
    """This object is the `root` of the command graph"""

    current_screen: Screen
    dgroups: DGroups
    _eventloop: asyncio.AbstractEventLoop

    def __init__(
        self,
        kore: base.Core,
        config: Config,
        no_spawn: bool = False,
        state: str | None = None,
        socket_path: str | None = None,
    ) -> None:
        self.core: base.Core = kore
        self.config = config
        self.no_spawn = no_spawn
        self._state: QtileState | str | None = state
        self.socket_path = socket_path

        self._drag: tuple | None = None
        self._mouse_map: defaultdict[int, list[Mouse]] = defaultdict(list)

        self.windows_map: dict[int, base.WindowType] = {}
        self.widgets_map: dict[str, _Widget] = {}
        self.renamed_widgets: list[str]
        self.groups_map: dict[str, _Group] = {}
        self.groups: list[_Group] = []
        self.hovered_window: base.WindowType | None = None

        self.keys_map: dict[tuple[int, int], Key | KeyChord] = {}
        self.chord_stack: list[KeyChord] = []

        self.screens: list[Screen] = []

        libqtile.init(self)
        libqtile.event_loop = asyncio.new_event_loop()

        self._stopped_event: asyncio.Event = asyncio.Event()

        self.server = IPCCommandServer(self)

        self.locked = False
        hook.subscribe.locked(self.lock)
        hook.subscribe.unlocked(self.unlock)
        self.test_data = None
        self.test_data_config_evaluations = None

    def lock(self) -> None:
        self.locked = True

    def unlock(self) -> None:
        self.locked = False

    def load_config(self, initial: bool = False) -> None:
        try:
            self.config.load()
            self.config.validate()
        except Exception as e:
            logger.exception("Configuration error:")
            send_notification("Configuration error", str(e))

        self.dgroups = DGroups(self, self.config.groups, self.config.dgroups_key_binder)

        _Widget.global_defaults = self.config.widget_defaults
        _Extension.global_defaults = self.config.extension_defaults

        for installed_extension in _Extension.installed_extensions:
            installed_extension._configure(self)

        for i in self.groups:
            self.groups_map[i.name] = i

        for grp in self.config.groups:
            if isinstance(grp, ScratchPadConfig):
                sp = ScratchPad(grp.name, grp.dropdowns, grp.label, grp.single)
                sp._configure([self.config.floating_layout], self.config.floating_layout, self)
                self.groups.append(sp)
                self.groups_map[sp.name] = sp

        self._process_screens(reloading=not initial)

        # Map and Grab keys
        for key in self.config.keys:
            self.grab_key(key)

        for button in self.config.mouse:
            self.grab_button(button)

        if self._state:
            if isinstance(self._state, str):
                try:
                    with open(self._state, "rb") as f:
                        st = pickle.load(f)
                        st.apply(self)
                except:  # noqa: E722
                    logger.exception("failed restoring state")
                finally:
                    os.remove(self._state)
            else:
                self._state.apply(self)

        self.core.on_config_load(initial)

        if self.config.reconfigure_screens:
            hook.subscribe.screen_change(self.reconfigure_screens)

        # no_spawn is set after the very first startup; we only want to run the
        # startup hook once. Note that this needs to happen *after* the config
        # is loaded and we have subscribed reconfigure_screens() in case people
        # do xrandr style manipulation in this hook.
        if not self.no_spawn:
            hook.fire("startup_once")
            self.no_spawn = True
        hook.fire("startup")

        if self._state:
            for screen in self.screens:
                screen.group.layout.show(screen.get_rect())
                screen.group.layout_all()
        self._state = None
        self.update_desktops()
        hook.subscribe.setgroup(self.update_desktops)

        # Start the sleep inhibitor process to listen to sleep signals
        # NB: the inhibitor will only connect to the dbus service if the
        # user has used the "suspend" or "resume" hooks in their config.
        inhibitor.start()

        if self.config.idle_inhibitors:
            self.core.idle_inhibitor_manager.set_hooks()

        self.core.idle_notifier.clear_timers()
        if self.config.idle_timers:
            self.core.idle_notifier.start()

        if initial:
            hook.fire("startup_complete")

    def _prepare_socket_path(
        self,
        socket_path: str | None = None,
    ) -> str:
        if socket_path is None:
            socket_path = ipc.find_sockfile(self.core.display_name)

        if os.path.exists(socket_path):
            os.unlink(socket_path)

        return socket_path

    def loop(self) -> None:
        asyncio.run(self.async_loop(), loop_factory=lambda: libqtile.event_loop)

    async def async_loop(self) -> None:
        """Run the event loop

        Finalizes the Qtile instance on exit.
        """
        self._eventloop = asyncio.get_running_loop()
        self.core.qtile = self
        self.load_config(initial=True)
        self.core.setup_listener()

        faulthandler.enable(all_threads=True)
        faulthandler.register(signal.SIGUSR2, all_threads=True)

        try:
            signals: dict[signal.Signals, Callable]
            signals = {
                signal.SIGTERM: self.stop,
                signal.SIGINT: self.stop,
                signal.SIGHUP: self.stop,
                signal.SIGUSR1: self.reload_config,
            }
            if self.core.name == "x11":
                # the wayland backend installs its own SIGCHLD handler after
                # the XWayland X server has initialized (as a workaround). the
                # x11 backend can just do it here.
                signals[signal.SIGCHLD] = utils.reap_zombies
            async with (
                LoopContext(signals),
                ipc.Server(
                    self._prepare_socket_path(self.socket_path),
                    self.server.call,
                ),
            ):
                await self._stopped_event.wait()
                if lifecycle.behavior != lifecycle.behavior.RESTART:
                    await self.graceful_shutdown()
        finally:
            self.finalize()
            self.core.remove_listener()

    def stop(self, exitcode: int = 0) -> None:
        hook.fire("shutdown")
        lifecycle.behavior = lifecycle.behavior.TERMINATE
        lifecycle.exitcode = exitcode
        self._stop()

    def normal_windows(self) -> list[base.WindowType]:
        return list(
            filter(lambda w: isinstance(w, base.Window), self.windows_map.copy().values())
        )

    async def graceful_shutdown(self) -> None:
        """Try to close windows gracefully before exiting"""
        # Copy in case the dictionary changes during the loop
        for win in self.normal_windows():
            win.kill()

        # give everyone a little time to exit and write their state. but don't
        # sleep forever (1s).
        end = time.time() + 1
        while time.time() < end:
            await asyncio.sleep(0.1)
            if len(self.normal_windows()) == 0:
                break

    @expose_command()
    def restart(self) -> None:
        """Restart Qtile.

        Can also be triggered by sending Qtile a SIGUSR2 signal.
        """
        if not self.core.supports_restarting:
            raise CommandError(f"Backend does not support restarting: {self.core.name}")
        try:
            self.config.load()
        except Exception as error:
            logger.exception("Preventing restart because of a configuration error:")
            send_notification("Configuration error", str(error.__context__))
            return

        hook.fire("restart")
        lifecycle.behavior = lifecycle.behavior.RESTART
        state_file = os.path.join(tempfile.gettempdir(), "qtile-state")
        with open(state_file, "wb") as f:
            self.dump_state(f)
        lifecycle.state_file = state_file
        self._stop()

    def _stop(self) -> None:
        logger.debug("Stopping qtile")
        self._stopped_event.set()

    def dump_state(self, buf: Any) -> None:
        try:
            pickle.dump(QtileState(self), buf, protocol=0)
        except:  # noqa: E722
            logger.exception("Unable to pickle qtile state")

    @expose_command()
    def reload_config(self) -> None:
        """
        Reload the configuration file.

        Can also be triggered by sending Qtile a SIGUSR1 signal.
        """
        logger.debug("Reloading the configuration file")

        try:
            self.config.load()
        except Exception as error:
            logger.exception("Configuration error:")
            send_notification("Configuration error", str(error))
            return

        self._state = QtileState(self, restart=False)
        self._finalize_configurables()
        hook.clear()
        self.ungrab_keys()
        self.chord_stack.clear()
        self.core.ungrab_buttons()
        self._mouse_map.clear()
        self.groups_map.clear()
        self.groups.clear()
        self.screens.clear()
        remove_dbus_rules()
        self.load_config()

    def _finalize_configurables(self) -> None:
        """
        Finalize objects that are instantiated within the config file. In addition to
        shutdown, these are finalized and then regenerated when reloading the config.
        """
        try:
            for widget in self.widgets_map.values():
                if not widget.finalized:
                    widget.finalize()
            self.widgets_map.clear()

            # For layouts we need to finalize each clone of a layout in each group
            for group in self.groups:
                for layout in group.layouts:
                    layout.finalize()

            for screen in self.screens:
                screen.finalize_gaps()

        except:  # noqa: E722
            logger.exception("exception during finalize")
        hook.clear()

    def finalize(self) -> None:
        self._finalize_configurables()
        remove_dbus_rules()
        inhibitor.stop()
        self.core.finalize()

    def add_autogen_group(self, screen_idx: int) -> _Group:
        name = f"autogen_{screen_idx + 1}"
        self.add_group(name)
        logger.warning("Too few groups in config. Added group: %s", name)
        return self.groups_map[name]

    def get_available_group(self, screen_idx: int) -> _Group | None:
        for group in self.groups:
            # Groups belonging to a screen or a scratchpad are not 'available'
            # to be assigned to a screen
            if group.screen or isinstance(group, ScratchPad):
                continue

            # Only return groups that can be tied to this screen
            # And thus do not have a "screen affinity" explicitly set for another screen
            if group.screen_affinity is None or group.screen_affinity == screen_idx:
                return group
        return None

    def get_output_info(self) -> list[Output]:
        if self.config.fake_screens is not None:
            return [
                Output(None, None, None, None, ScreenRect(s.x, s.y, s.width, s.height))
                for s in self.config.fake_screens
            ]

        # Alias screens with the same x and y coordinates, taking largest
        xywh: dict[
            tuple[int, int], tuple[int, int, str | None, str | None, str | None, str | None]
        ] = {}
        for info in self.core.get_output_info():
            pos = (info.rect.x, info.rect.y)
            width, height, port, make, model, serial = xywh.get(
                pos, (0, 0, info.port, info.make, info.model, info.serial)
            )
            # if one monitor is wider and one monitor is longer, either
            # serial number was valid (i.e. we could choose either, since
            # we're going to project over the whole space). just pick one.
            xywh[pos] = (
                max(width, info.rect.width),
                max(height, info.rect.height),
                info.port,
                info.make,
                info.model,
                info.serial,
            )

        return [
            Output(port, make, model, serial, ScreenRect(x, y, w, h))
            for (x, y), (w, h, port, make, model, serial) in xywh.items()
        ]

    def get_screens_from_config(self, output_info: list[Output]) -> list[Screen]:
        if self.config.fake_screens is not None:
            return self.config.fake_screens

        if self.config.generate_screens is not None:
            if self.config.screens is not None:
                logger.warning(
                    "Both screens and generate_screens are defined in config. Using generate_screens."
                )
            return self.config.generate_screens(output_info)

        return self.config.screens

    def _process_screens(self, reloading: bool = False) -> None:
        current_groups = [s.group for s in self.screens]
        output_info = self.get_output_info()
        config_screens = self.get_screens_from_config(output_info)
        new_screens: list[Screen] = []

        for i, info in enumerate(output_info):
            if i < len(config_screens):
                scr = config_screens[i]
                logger.debug(f"using config at index {i} for output {info.port}")
            else:
                # user didn't supply enough screens, create one
                scr = Screen()
            scr.output = info

            if not hasattr(self, "current_screen") or reloading:
                self.current_screen = scr
                reloading = False

            grp = None
            if i < len(current_groups):
                grp = current_groups[i]
            else:
                # We need to assign a new group
                # Get an available group or create a new one
                grp = self.get_available_group(i)
                if grp is None:
                    grp = self.add_autogen_group(i)

            # If the screen has changed position and/or size, or is a new screen then make sure that any gaps/bars
            # are reconfigured
            reconfigure_gaps = (
                (info.rect.x, info.rect.y, info.rect.width, info.rect.height)
                != (scr.x, scr.y, scr.width, scr.height)
            ) or (i + 1 > len(self.screens))

            if not hasattr(scr, "group"):
                # Ensure that this screen actually *has* a group, as it won't get
                # assigned one during `__init__` because they are created in the config,
                # where the groups also are. This lets us type `Screen.group` as
                # `_Group` rather than `_Group | None` which would need lots of other
                # changes to check for `None`s, and conceptually all screens should have
                # a group anyway.
                scr.group = grp

            scr._configure(
                self,
                i,
                info.rect.x,
                info.rect.y,
                info.rect.width,
                info.rect.height,
                grp,
                reconfigure_gaps=reconfigure_gaps,
            )
            new_screens.append(scr)

        for screen in self.screens:
            if screen not in new_screens:
                screen.finalize_gaps()

        self.screens = new_screens

    @expose_command()
    def reconfigure_screens(self, *_: list[Any], **__: dict[Any, Any]) -> None:
        """
        This can be used to set up screens again during run time. Intended usage is to
        be called when the screen_change hook is fired, responding to changes in
        physical monitor setup by configuring qtile.screens accordingly. The args are
        ignored; it is here in case this function is hooked directly to screen_change.
        """
        logger.info("Reconfiguring screens.")
        self._process_screens()

        for group in self.groups:
            if group.screen:
                if group.screen in self.screens:
                    group.layout_all()
                else:
                    group.hide()

        hook.fire("screens_reconfigured")

    def paint_screen(self, screen: Screen, image_path: str, mode: str | None = None) -> None:
        self.core.painter.paint(screen, image_path, mode)

    def fill_screen(self, screen: Screen, background: ColorType) -> None:
        self.core.painter.fill(screen, background)

    def process_key_event(self, keysym: int, mask: int) -> tuple[Key | KeyChord | None, bool]:
        key = self.keys_map.get((keysym, mask), None)
        if key is None:
            logger.debug("Ignoring unknown keysym: %s, mask: %s", keysym, mask)
            return (None, False)

        if isinstance(key, KeyChord):
            self.grab_chord(key)
        else:
            # Keep track if we have executed a command
            executed = False
            for cmd in key.commands:
                if cmd.check(self):
                    status, val = self.server.call(
                        (cmd.selectors, cmd.name, cmd.args, cmd.kwargs, False)
                    )
                    if status in (interface.ERROR, interface.EXCEPTION):
                        logger.error("KB command error %s: %s", cmd.name, val)
                    executed = True
            if self.chord_stack and (not self.chord_stack[-1].mode or key.key == "Escape"):
                self.ungrab_chord()
            # We never swallow when no commands have been executed,
            # even when key.swallow is set to True
            elif not executed:
                return (key, False)
        # Return whether we have handled the key based on the key's swallow parameter
        return (key, key.swallow)

    def grab_keys(self) -> None:
        """Re-grab all of the keys configured in the key map

        Useful when a keyboard mapping event is received.
        """
        self.core.ungrab_keys()
        keys = self.keys_map.copy()
        self.keys_map.clear()
        for key in keys.values():
            self.grab_key(key)

    def grab_key(self, key: Key | KeyChord) -> None:
        """Grab the given key event"""
        syms = self.core.grab_key(key)
        if syms in self.keys_map:
            if self.keys_map[syms] == key:
                # We've already bound this key definition
                return
            logger.warning("Key spec duplicated, overriding previous: %s", key)
        self.keys_map[syms] = key

    def ungrab_key(self, key: Key | KeyChord) -> None:
        """Ungrab a given key event"""
        keysym, mask_key = self.core.ungrab_key(key)
        self.keys_map.pop((keysym, mask_key))

    def ungrab_keys(self) -> None:
        """Ungrab all key events"""
        self.core.ungrab_keys()
        self.keys_map.clear()

    def grab_chord(self, chord: KeyChord) -> None:
        self.chord_stack.append(chord)
        if self.chord_stack:
            hook.fire("enter_chord", chord.name)

        self.ungrab_keys()
        for key in chord.submappings:
            self.grab_key(key)

    @expose_command()
    def ungrab_chord(self) -> None:
        """Leave a chord mode"""
        hook.fire("leave_chord")

        self.ungrab_keys()
        if not self.chord_stack:
            logger.debug("ungrab_chord was called when no chord mode was active")
            return
        # The first pop is necessary: Otherwise we would be stuck in a mode;
        # we could not leave it: the code below would re-enter the old mode.
        self.chord_stack.pop()
        # Find another named mode or load the root keybindings:
        while self.chord_stack:
            chord = self.chord_stack.pop()
            if chord.mode:
                self.grab_chord(chord)
                break
        else:
            for key in self.config.keys:
                self.grab_key(key)

    @expose_command()
    def ungrab_all_chords(self) -> None:
        """Leave all chord modes and grab the root bindings"""
        hook.fire("leave_chord")
        self.ungrab_keys()
        self.chord_stack.clear()
        for key in self.config.keys:
            self.grab_key(key)

    def grab_button(self, button: Mouse) -> None:
        """Grab the given mouse button event"""
        try:
            button.modmask = self.core.grab_button(button)
        except utils.QtileError:
            logger.warning("Unknown modifier(s): %s", button.modifiers)
            return
        self._mouse_map[button.button_code].append(button)

    def update_desktops(self) -> None:
        try:
            index = self.groups.index(self.current_group)
        # TODO: we should really only except ValueError here, AttributeError is
        # an annoying chicken and egg because we're accessing current_screen
        # (via current_group), and when we set up the initial groups, there
        # aren't any screens yet. This can probably be changed when #475 is
        # fixed.
        except (ValueError, AttributeError):
            index = 0

        self.core.update_desktops(self.groups, index)

    def add_group(
        self,
        name: str,
        layout: str | None = None,
        layouts: list[Layout] | None = None,
        label: str | None = None,
        index: int | None = None,
        screen_affinity: int | None = None,
        persist: bool | None = False,
    ) -> bool:
        if name not in self.groups_map.keys():
            g = _Group(
                name, layout, label=label, screen_affinity=screen_affinity, persist=persist
            )
            if index is None:
                self.groups.append(g)
            else:
                self.groups.insert(index, g)

            if not layouts:
                layouts = self.config.layouts
            g._configure(layouts, self.config.floating_layout, self)
            self.groups_map[name] = g
            hook.fire("addgroup", name)
            hook.fire("changegroup")
            self.update_desktops()

            return True
        return False

    def delete_group(self, name: str) -> None:
        # one group per screen is needed
        if len(self.groups) == len(self.screens):
            raise ValueError("Can't delete all groups.")

        if name in self.groups_map.keys():
            group = self.groups_map[name]

            # Find a group that's not currently on a screen to bring to the front.
            target = group.get_previous_group(skip_managed=True)

            # move windows to other group
            for i in list(group.windows):
                i.togroup(target.name)

            # if group to be deleted is currently active
            if self.current_group.name == name:
                # switch to target group
                self.current_screen.set_group(target, save_prev=False)

            self.groups.remove(group)
            del self.groups_map[name]

            hook.fire("delgroup", name)
            hook.fire("changegroup")

            self.update_desktops()

    def register_widget(self, w: _Widget) -> None:
        """
        Register a bar widget

        If a widget with the same name already exists, the new widget will be
        automatically renamed by appending numeric suffixes. For example, if
        the widget is named "foo", we will attempt "foo_1", "foo_2", and so on,
        until a free name is found.

        This naming convention is only used for qtile.widgets_map as every widget
        MUST be registered here to ensure that objects are finalised correctly.

        Widgets can still be accessed by their name when using
        lazy.screen.widget[name] or lazy.bar["top"].widget[name] unless there are
        duplicate widgets in the bar/screen.

        A warning will be provided where renaming has occurred.
        """
        # Find unoccupied name by appending numeric suffixes
        name = w.name
        i = 0
        while name in self.widgets_map:
            i += 1
            name = f"{w.name}_{i}"

        if name != w.name:
            self.renamed_widgets.append(name)

        self.widgets_map[name] = w

    @property
    def current_layout(self) -> Layout:
        return self.current_group.layout

    @property
    def current_group(self) -> _Group:
        return self.current_screen.group

    @property
    def current_window(self) -> base.Window | None:
        return self.current_screen.group.current_window

    def reserve_space(
        self,
        reserved_space: tuple[int, int, int, int],  # [left, right, top, bottom]
        screen: Screen,
    ) -> None:
        """
        Reserve some space at the edge(s) of a screen.

        The requested space is added to space reserved previously: repeated calls to
        this method are not idempotent.
        """
        for i, pos in enumerate(["left", "right", "top", "bottom"]):
            if space := reserved_space[i]:
                if gap := getattr(screen, pos):
                    gap.adjust_reserved_space(space)
                elif 0 < space:
                    gap = bar.Gap(0)
                    gap.screen = screen
                    setattr(screen, pos, gap)
                    gap.adjust_reserved_space(space)
        screen.resize()

    def free_reserved_space(
        self,
        reserved_space: tuple[int, int, int, int],  # [left, right, top, bottom]
        screen: Screen,
    ) -> None:
        """
        Free up space that has previously been reserved at the edge(s) of a screen.
        """
        # mypy can't work out that the new tuple is also length 4 (see mypy #7509)
        reserved_space = tuple(-i for i in reserved_space)  # type: ignore
        self.reserve_space(reserved_space, screen)

    def manage(self, win: base.WindowType) -> None:
        if isinstance(win, base.Internal):
            self.windows_map[win.wid] = win
            return

        if win.wid in self.windows_map:
            return

        hook.fire("client_new", win)

        # Window may be defunct because
        # it's been declared static in hook.
        if win.defunct:
            return
        self.windows_map[win.wid] = win
        if self.current_screen and isinstance(win, base.Window):
            # Window may have been bound to a group in the hook.
            if not win.group and self.current_screen.group:
                self.current_screen.group.add(win)

            # Check if any user-defined inhibitor rules match the window
            win.add_config_inhibitors()

        hook.fire("client_managed", win)

    def unmanage(self, wid: int) -> None:
        c = self.windows_map.get(wid)
        if c:
            # Fire the hook before removing the group from the window so hooked
            # functions can access it
            hook.fire("client_killed", c)
            if isinstance(c, base.Static):
                if c.reserved_space:
                    self.free_reserved_space(c.reserved_space, c.screen)
            elif isinstance(c, base.Window):
                if c.group:
                    c.group.remove(c)
            del self.windows_map[wid]
            if isinstance(c, base.Window):
                self.core.idle_inhibitor_manager.remove_window_inhibitor(c)

    def find_screen(self, x: int, y: int) -> Screen | None:
        """Find a screen based on the x and y offset"""
        result = []
        for i in self.screens:
            if i.x <= x <= i.x + i.width and i.y <= y <= i.y + i.height:
                result.append(i)
        if len(result) == 1:
            return result[0]
        return None

    def find_closest_screen(self, x: int, y: int) -> Screen:
        """
        If find_screen returns None, then this basically extends a
        screen vertically and horizontally and see if x,y lies in the
        band.

        Only works if it can find a SINGLE closest screen, else we
        revert to _find_closest_closest.

        Useful when dragging a window out of a screen onto another but
        having leftmost corner above viewport.
        """
        normal = self.find_screen(x, y)
        if normal is not None:
            return normal
        x_match = []
        y_match = []
        for i in self.screens:
            if i.x <= x <= i.x + i.width:
                x_match.append(i)
            if i.y <= y <= i.y + i.height:
                y_match.append(i)
        if len(x_match) == 1:
            return x_match[0]
        if len(y_match) == 1:
            return y_match[0]
        return self._find_closest_closest(x, y, x_match + y_match)

    def _find_closest_closest(self, x: int, y: int, candidate_screens: list[Screen]) -> Screen:
        """
        if find_closest_screen can't determine one, we've got multiple
        screens, so figure out who is closer.  We'll calculate using
        the square of the distance from the center of a screen.

        Note that this could return None if x, y is right/below all
        screens.
        """
        closest_distance: float | None = None
        if not candidate_screens:
            # try all screens
            candidate_screens = self.screens
        # if left corner is below and right of screen
        # it can't really be a candidate
        candidate_screens = [
            s for s in candidate_screens if x < s.x + s.width and y < s.y + s.height
        ]
        closest_screen = lget(candidate_screens, 0)
        for s in candidate_screens:
            middle_x = s.x + s.width / 2
            middle_y = s.y + s.height / 2
            distance = (x - middle_x) ** 2 + (y - middle_y) ** 2
            if closest_distance is None or distance < closest_distance:
                closest_distance = distance
                closest_screen = s
        return closest_screen or self.screens[0]

    def focus_hovered_window(self) -> None:
        window = self.hovered_window
        if window:
            if isinstance(window, base.Window):
                window.focus()

    def process_button_click(self, button_code: int, modmask: int, x: int, y: int) -> bool:
        handled = False
        for m in self._mouse_map[button_code]:
            if not m.modmask == modmask:
                continue

            if isinstance(m, Click):
                if self.config.follow_mouse_focus == "click_or_drag_only":
                    self.focus_hovered_window()
                for i in m.commands:
                    if i.check(self):
                        status, val = self.server.call(
                            (i.selectors, i.name, i.args, i.kwargs, False)
                        )
                        if status in (interface.ERROR, interface.EXCEPTION):
                            logger.error("Mouse command error %s: %s", i.name, val)
                        handled = True
            elif (
                isinstance(m, Drag) and self.current_window and not self.current_window.fullscreen
            ):
                if self.config.follow_mouse_focus == "click_or_drag_only":
                    self.focus_hovered_window()
                if m.start:
                    i = m.start
                    status, val = self.server.call((i.selectors, i.name, i.args, i.kwargs, False))
                    if status in (interface.ERROR, interface.EXCEPTION):
                        logger.error("Mouse command error %s: %s", i.name, val)
                        continue
                else:
                    val = (0, 0)

                if m.warp_pointer and self.current_window is not None:
                    win_size = self.current_window.get_size()
                    win_pos = self.current_window.get_position()
                    x = win_size[0] + win_pos[0]
                    y = win_size[1] + win_pos[1]
                    self.core.warp_pointer(x, y)

                self._drag = (x, y, val[0], val[1], m.commands)
                self.core.grab_pointer()
                handled = True

        return handled

    def process_button_release(self, button_code: int, modmask: int) -> bool:
        if self._drag is not None:
            for m in self._mouse_map[button_code]:
                if isinstance(m, Drag):
                    self._drag = None
                    self.core.ungrab_pointer()
                    return True
        return False

    def process_button_motion(self, x: int, y: int) -> None:
        if self._drag is None:
            return
        ox, oy, rx, ry, cmd = self._drag
        dx = x - ox
        dy = y - oy
        if dx or dy:
            for i in cmd:
                if i.check(self):
                    status, val = self.server.call(
                        (i.selectors, i.name, i.args + (rx + dx, ry + dy), i.kwargs, False)
                    )
                    if status in (interface.ERROR, interface.EXCEPTION):
                        logger.error("Mouse command error %s: %s", i.name, val)

    def warp_to_screen(self) -> None:
        if self.current_screen:
            scr = self.current_screen
            self.core.warp_pointer(scr.x + scr.dwidth // 2, scr.y + scr.dheight // 2)

    def focus_screen(self, n: int, warp: bool = True) -> None:
        """Have Qtile move to screen and put focus there"""
        if n >= len(self.screens):
            return
        old = self.current_screen
        self.current_screen = self.screens[n]
        if old != self.current_screen:
            hook.fire("current_screen_change")
            hook.fire("setgroup")
            old.group.layout_all()
            self.current_group.focus(self.current_window, warp)
            if self.current_window is None and warp:
                self.warp_to_screen()

    def move_to_group(self, group: str) -> None:
        """Create a group if it doesn't exist and move
        the current window there"""
        if self.current_window and group:
            self.add_group(group)
            self.current_window.togroup(group)

    def _items(self, name: str) -> ItemT:
        if name == "group":
            return True, list(self.groups_map.keys())
        elif name == "layout":
            return True, list(range(len(self.current_group.layouts)))
        elif name == "widget":
            return False, list(self.widgets_map.keys())
        elif name == "bar":
            return False, [x.position for x in self.current_screen.gaps if isinstance(x, bar.Bar)]
        elif name == "window":
            windows: list[str | int]
            windows = [
                k
                for k, v in self.windows_map.items()
                if isinstance(v, CommandObject) and not isinstance(v, _Widget)
            ]
            return True, windows
        elif name == "screen":
            return True, list(range(len(self.screens)))
        elif name == "core":
            return True, []
        return None

    def _select(self, name: str, sel: str | int | None) -> CommandObject | None:
        if name == "group":
            if sel is None:
                return self.current_group
            else:
                return self.groups_map.get(sel)  # type: ignore
        elif name == "layout":
            if sel is None:
                return self.current_group.layout
            else:
                return lget(self.current_group.layouts, int(sel))
        elif name == "widget":
            return self.widgets_map.get(sel)  # type: ignore
        elif name == "bar":
            gap = getattr(self.current_screen, sel)  # type: ignore
            if isinstance(gap, bar.Bar):
                return gap
        elif name == "window":
            if sel is None:
                return self.current_window
            else:
                windows: dict[str | int, base.WindowType]
                windows = {
                    k: v
                    for k, v in self.windows_map.items()
                    if isinstance(v, CommandObject) and not isinstance(v, _Widget)
                }
                return windows.get(sel)
        elif name == "screen":
            if sel is None:
                return self.current_screen
            else:
                return lget(self.screens, int(sel))
        elif name == "core":
            return self.core
        return None

    def call_soon(self, func: Callable, *args: Any) -> asyncio.Handle:
        """A wrapper for the event loop's call_soon which also flushes the core's
        event queue after func is called."""

        def f() -> None:
            func(*args)
            self.core.flush()

        return self._eventloop.call_soon(f)

    def call_soon_threadsafe(self, func: Callable, *args: Any) -> asyncio.Handle:
        """Another event loop proxy, see `call_soon`."""

        def f() -> None:
            func(*args)
            self.core.flush()

        return self._eventloop.call_soon_threadsafe(f)

    def call_later(self, delay: int | float, func: Callable, *args: Any) -> asyncio.TimerHandle:
        """Another event loop proxy, see `call_soon`."""

        def f() -> None:
            func(*args)
            self.core.flush()

        return self._eventloop.call_later(delay, f)

    def run_in_executor(self, func: Callable, *args: Any) -> asyncio.Future:
        """A wrapper for running a function in the event loop's default
        executor."""
        return self._eventloop.run_in_executor(None, func, *args)

    @expose_command()
    def debug(self) -> None:
        """Set log level to DEBUG"""
        logger.setLevel(logging.DEBUG)
        self.core.update_backend_log_level()
        logger.debug("Switching to DEBUG threshold")

    @expose_command()
    def info(self) -> None:
        """Set log level to INFO"""
        logger.setLevel(logging.INFO)
        self.core.update_backend_log_level()
        logger.info("Switching to INFO threshold")

    @expose_command()
    def warning(self) -> None:
        """Set log level to WARNING"""
        logger.setLevel(logging.WARNING)
        self.core.update_backend_log_level()
        logger.warning("Switching to WARNING threshold")

    @expose_command()
    def error(self) -> None:
        """Set log level to ERROR"""
        logger.setLevel(logging.ERROR)
        self.core.update_backend_log_level()
        logger.error("Switching to ERROR threshold")

    @expose_command()
    def critical(self) -> None:
        """Set log level to CRITICAL"""
        logger.setLevel(logging.CRITICAL)
        self.core.update_backend_log_level()
        logger.critical("Switching to CRITICAL threshold")

    @expose_command()
    def loglevel(self) -> int:
        return logger.level

    @expose_command()
    def loglevelname(self) -> str:
        return logging.getLevelName(logger.level)

    @expose_command()
    def pause(self) -> None:
        """Drops into pdb"""
        import pdb

        pdb.set_trace()

    @expose_command()
    def get_groups(self) -> dict[str, dict[str, Any]]:
        """
        Return a dictionary containing information for all groups

        Examples
        ========

            get_groups()
        """
        return {i.name: i.info() for i in self.groups}

    @expose_command()
    def display_kb(self) -> str:
        """Display table of key bindings"""

        class FormatTable:
            def __init__(self) -> None:
                self.max_col_size: list[int] = []
                self.rows: list[list[str]] = []

            def add(self, row: list[str]) -> None:
                n = len(row) - len(self.max_col_size)
                if n > 0:
                    self.max_col_size += [0] * n
                for i, f in enumerate(row):
                    if len(f) > self.max_col_size[i]:
                        self.max_col_size[i] = len(f)
                self.rows.append(row)

            def getformat(self) -> tuple[str, int]:
                format_string = " ".join(
                    f"%-{max_col_size + 2:d}s" for max_col_size in self.max_col_size
                )
                return format_string + "\n", len(self.max_col_size)

            def expandlist(self, list_: list[str], n: int) -> list[str]:
                if not list_:
                    return ["-" * max_col_size for max_col_size in self.max_col_size]
                n -= len(list_)
                if n > 0:
                    list_ += [""] * n
                return list_

            def __str__(self) -> str:
                format_, n = self.getformat()
                return "".join(format_ % tuple(self.expandlist(row, n)) for row in self.rows)

        result = FormatTable()
        result.add(["Mode", "KeySym", "Mod", "Command", "Desc"])
        result.add([])
        rows = []

        def walk_binding(k: Key | KeyChord, mode: str) -> None:
            nonlocal rows
            modifiers = ", ".join(k.modifiers)
            if isinstance(k.key, int):
                name = hex(k.key)
            else:
                name = k.key

            if isinstance(k, Key):
                if not k.commands:
                    return
                allargs = ", ".join(
                    [
                        value.__name__ if callable(value) else repr(value)
                        for value in k.commands[0].args
                    ]
                    + [
                        f"{keyword} = {repr(value)}"
                        for keyword, value in k.commands[0].kwargs.items()
                    ]
                )
                rows.append(
                    [
                        mode,
                        name,
                        modifiers,
                        f"{k.commands[0].name:s}({allargs:s})",
                        k.desc,
                    ]
                )
                return
            if isinstance(k, KeyChord):
                new_mode_s = k.name if k.name else "<unnamed>"
                new_mode = (
                    k.name
                    if mode == "<root>"
                    else "{}>{}".format(mode, k.name if k.name else "_")
                )
                rows.append([mode, name, modifiers, "", f"Enter {new_mode_s:s} mode"])
                for s in k.submappings:
                    walk_binding(s, new_mode)
                return
            raise TypeError(f"Unexpected type: {type(k)}")

        for k in self.config.keys:
            walk_binding(k, "<root>")
        rows.sort()
        for row in rows:
            result.add(row)
        return str(result)

    @expose_command()
    def list_widgets(self) -> list[str]:
        """List of all addressible widget names"""
        return list(self.widgets_map.keys())

    @expose_command()
    def to_layout_index(self, index: int, name: str | None = None) -> None:
        """
        Switch to the layout with the given index in self.layouts.

        Parameters
        ==========
        index :
            Index of the layout in the list of layouts.
        name :
            Group name. If not specified, the current group is assumed.
        """
        if name is not None:
            group = self.groups_map[name]
        else:
            group = self.current_group
        group.use_layout(index)

    @expose_command()
    def next_layout(self, name: str | None = None) -> None:
        """
        Switch to the next layout.

        Parameters
        ==========
        name :
            Group name. If not specified, the current group is assumed
        """
        if name is not None:
            group = self.groups_map[name]
        else:
            group = self.current_group
        group.use_next_layout()

    @expose_command()
    def prev_layout(self, name: str | None = None) -> None:
        """
        Switch to the previous layout.

        Parameters
        ==========
        name :
            Group name. If not specified, the current group is assumed
        """
        if name is not None:
            group = self.groups_map[name]
        else:
            group = self.current_group
        group.use_previous_layout()

    @expose_command()
    def get_screens(self) -> list[dict[str, Any]]:
        """Return a list of dictionaries providing information on all screens"""
        lst = [
            dict(
                index=i.index,
                group=i.group.name if i.group is not None else None,
                x=i.x,
                y=i.y,
                width=i.width,
                height=i.height,
                gaps=dict(
                    top=i.top.geometry() if i.top else None,
                    bottom=i.bottom.geometry() if i.bottom else None,
                    left=i.left.geometry() if i.left else None,
                    right=i.right.geometry() if i.right else None,
                ),
            )
            for i in self.screens
        ]
        return lst

    @expose_command()
    def simulate_keypress(self, modifiers: list[str], key: str) -> None:
        """Simulates a keypress on the focused window.

        This triggers internal bindings only; for full simulation see external tools
        such as xdotool or ydotool.

        Parameters
        ==========
        modifiers :
            A list of modifier specification strings. Modifiers can be one of
            "shift", "lock", "control" and "mod1" - "mod5".
        key :
            Key specification.

        Examples
        ========
            simulate_keypress(["control", "mod2"], "k")
        """
        try:
            self.core.simulate_keypress(modifiers, key)
        except utils.QtileError as e:
            raise CommandError(str(e))

    @expose_command()
    def validate_config(self) -> None:
        try:
            self.config.load()
        except Exception as error:
            send_notification("Configuration check", str(error))
        else:
            send_notification("Configuration check", "No error found!")

    @expose_command()
    def spawn(
        self,
        cmd: list[str] | str | Sequence[str | PathLike],
        shell: bool = False,
        env: dict[str, str] = dict(),
        group: str | None = None,
    ) -> int:
        """
        Spawn a new process.

        Parameters
        ==========
        cmd:
            The command to execute either as a single string or list of strings.
        shell:
            Whether to execute the command in a new shell by prepending it with "/bin/sh
            -c". This enables the use of shell syntax within the command (e.g. pipes).
        env:
            Dictionary of environmental variables to pass with command.
        group:
            Name of group in which to spawn process.

        Examples
        ========

            spawn("firefox")

            spawn(["xterm", "-T", "Temporary terminal"])

            spawn("screenshot | xclip", shell=True)
        """

        def expand(arg: str | PathLike) -> str:
            if isinstance(arg, Path):
                arg = arg.expanduser()
            return os.fspath(arg)

        if isinstance(cmd, str | os.PathLike):
            args = shlex.split(expand(cmd))
        else:
            args = [expand(c) for c in cmd]
            cmd = shlex.join(args)

        to_lookup = args[0]
        if shell:
            args = ["/bin/sh", "-c", cmd]

        if shutil.which(to_lookup) is None:
            logger.error("couldn't find `%s`", to_lookup)
            return -1

        if len(env) == 0:
            env = os.environ.copy()
            # if qtile was installed in a virutal env, we don't
            # necessarily want to propagate that to children
            # applications, since it may change e.g. the behavior
            # of shells that spawn python applications
            env.pop("VIRTUAL_ENV", None)

        # std{in,out,err} should be /dev/null
        with open("/dev/null") as null:
            file_actions: list[tuple] = [
                (os.POSIX_SPAWN_DUP2, null.fileno(), 0),
                (os.POSIX_SPAWN_DUP2, 1, null.fileno()),
                (os.POSIX_SPAWN_DUP2, 2, null.fileno()),
            ]

            # To create a rule to move process to a specific group and manage race conditions
            # we need to do the following:
            # 1) Create a socket pair so we can communicate to child process
            # 2) Spawn a "qtile launch" process and pass the file descriptor
            # 3) Have "qtile launch" notify qtile that it's ready
            # 4) Create a rule using the pid from step 2
            # 5) Send a message over the socket to "qtile launch"
            # 6) Once the message is received, "qtile launch" execs the desired process
            if group is not None:
                parent_sock, child_sock = socket.socketpair()
                # socket pair fds are non-inheritable by default but we need this to survice
                os.set_inheritable(child_sock.fileno(), True)

            # this API is only available on python 3.13 or above, and only on platforms
            # where posix_spawn_file_actions_addclosefrom_np() exists (only glibc on Linux).
            # where we're creating a rule to manage the group, we'll handle closing fds in
            # `qtile launch`
            if hasattr(os, "POSIX_SPAWN_CLOSEFROM") and group is None:
                # we should close all fds so that child processes don't
                # accidentally write to our x11 event loop or whatever; we never
                # used to do this, so it seems fine to only do it where this nice
                # API to do it exists.
                file_actions.append((os.POSIX_SPAWN_CLOSEFROM, 3))

            try:
                if group is not None:
                    # As per step 2 (above) we want to launch the process via "qtile launch"
                    args = ["qtile", "launch", "--fd", f"{child_sock.fileno()}"] + args
                pid = os.posix_spawnp(args[0], args, env, file_actions=file_actions)
                if group is not None:
                    # Listen for READY message from `qtile launch`
                    data = parent_sock.recv(5)

                    if data != b"READY":
                        logger.warning("Received unexpected data: {data!r}")

                    # Create the group matching rule
                    match_args = {"net_wm_pid": pid}
                    rule_args = {"group": group, "one_time": True}
                    self.add_rule(match_args=match_args, rule_args=rule_args)

                    # Tell child process it can now spawn main proces as rule has been added
                    parent_sock.send(b"OK")

                    parent_sock.close()
                    child_sock.close()
                return pid
            except OSError as e:
                logger.warning("failed to execute: %s: %s", str(args), str(e))
                return -1

    @expose_command()
    def status(self) -> Literal["OK"]:
        """Return "OK" if Qtile is running"""
        return "OK"

    @expose_command()
    def sync(self) -> None:
        """
        Sync the backend's event queue. Should only be used for development.
        """
        self.core.flush()

    @expose_command()
    def to_screen(self, n: int) -> None:
        """Warp focus to screen n, where n is a 0-based screen number

        Examples
        ========

            to_screen(0)
        """
        self.focus_screen(n)

    @expose_command()
    def next_screen(self) -> None:
        """Move to next screen"""
        self.focus_screen((self.screens.index(self.current_screen) + 1) % len(self.screens))

    @expose_command()
    def prev_screen(self) -> None:
        """Move to the previous screen"""
        self.focus_screen((self.screens.index(self.current_screen) - 1) % len(self.screens))

    @expose_command()
    def windows(self) -> list[dict[str, Any]]:
        """Return info for each client window"""
        return [
            i.info()
            for i in self.windows_map.values()
            if not isinstance(i, base.Internal | _Widget) and isinstance(i, CommandObject)
        ]

    def lookup_client(self, wid: int) -> base.Window | None:
        w = self.windows_map.get(wid)
        if isinstance(w, base.Window):
            return w
        return None

    @expose_command()
    def internal_windows(self) -> list[dict[str, Any]]:
        """Return info for each internal window (bars, for example)"""
        return [i.info() for i in self.windows_map.values() if isinstance(i, base.Internal)]

    @expose_command()
    def qtile_info(self) -> dict:
        """Returns a dictionary of info on the Qtile instance"""
        config_path = self.config.file_path
        dictionary = {
            "version": VERSION,
            "log_level": self.loglevelname(),
        }

        if isinstance(logger.handlers[0], RotatingFileHandler):
            log_path = logger.handlers[0].baseFilename
            dictionary["log_path"] = log_path

        if isinstance(config_path, str):
            dictionary["config_path"] = config_path
        elif isinstance(config_path, Path):
            dictionary["config_path"] = config_path.as_posix()

        return dictionary

    @expose_command()
    def shutdown(self, exitcode: int = 0) -> None:
        """Quit Qtile

        Parameters
        ==========
        exitcode :
            Set exit status of Qtile. Can be e.g. used to make login managers
            poweroff or restart the system. (default: 0)
        """
        self.stop(exitcode)

    @expose_command()
    def switch_groups(self, namea: str, nameb: str) -> None:
        """Switch position of two groups by name"""
        if namea not in self.groups_map or nameb not in self.groups_map:
            return

        indexa = self.groups.index(self.groups_map[namea])
        indexb = self.groups.index(self.groups_map[nameb])

        self.groups[indexa], self.groups[indexb] = self.groups[indexb], self.groups[indexa]
        hook.fire("setgroup")

        # update window _NET_WM_DESKTOP
        for group in (self.groups[indexa], self.groups[indexb]):
            for w in group.windows:
                w.group = group

    def find_window(self, wid: int) -> None:
        window = self.windows_map.get(wid)
        if isinstance(window, base.Window) and window.group:
            if not window.group.screen:
                self.current_screen.set_group(window.group)
            window.group.focus(window, False)

    @expose_command()
    def findwindow(self, prompt: str = "window", widget: str = "prompt") -> None:
        """Launch prompt widget to find a window of the given name

        Parameters
        ==========
        prompt :
            Text with which to prompt user (default: "window")
        widget :
            Name of the prompt widget (default: "prompt")
        """
        mb = self.widgets_map.get(widget)
        if not mb:
            logger.error("No widget named '%s' present.", widget)
            return

        mb.start_input(prompt, self.find_window, "window", strict_completer=True)

    @expose_command()
    def switch_window(self, location: int) -> None:
        """
        Change to the window at the specified index in the current group.
        """
        windows = self.current_group.windows
        if location < 1 or location > len(windows):
            return

        self.current_group.focus(windows[location - 1])

    @expose_command()
    def change_window_order(self, new_location: int) -> None:
        """
        Change the order of the current window within the current group.
        """
        if new_location < 1 or new_location > len(self.current_group.windows):
            return

        windows = self.current_group.windows
        current_window_index = windows.index(self.current_window)

        temp = windows[current_window_index]
        windows[current_window_index] = windows[new_location - 1]
        windows[new_location - 1] = temp

    @expose_command()
    def next_urgent(self) -> None:
        """Focus next window with urgent hint"""
        try:
            nxt = [w for w in self.windows_map.values() if w.urgent][0]
            assert isinstance(nxt, base.Window)
            if nxt.group:
                nxt.group.toscreen()
                nxt.group.focus(nxt)
            else:
                self.current_screen.group.add(nxt)
                self.current_screen.group.focus(nxt)
        except IndexError:
            pass  # no window had urgent set

    @expose_command()
    def togroup(self, prompt: str = "group", widget: str = "prompt") -> None:
        """Launch prompt widget to move current window to a given group

        Parameters
        ==========
        prompt :
            Text with which to prompt user (default: "group")
        widget :
            Name of the prompt widget (default: "prompt")
        """
        if not self.current_window:
            logger.warning("No window to move")
            return

        mb = self.widgets_map.get(widget)
        if not mb:
            logger.error("No widget named '%s' present.", widget)
            return

        mb.start_input(prompt, self.move_to_group, "group", strict_completer=True)

    @expose_command()
    def switchgroup(self, prompt: str = "group", widget: str = "prompt") -> None:
        """Launch prompt widget to switch to a given group to the current screen

        Parameters
        ==========
        prompt :
            Text with which to prompt user (default: "group")
        widget :
            Name of the prompt widget (default: "prompt")
        """

        def f(group: str) -> None:
            if group:
                try:
                    self.groups_map[group].toscreen()
                except KeyError:
                    logger.warning("No group named '%s' present.", group)

        mb = self.widgets_map.get(widget)
        if not mb:
            logger.error("No widget named '%s' present.", widget)
            return

        mb.start_input(prompt, f, "group", strict_completer=True)

    @expose_command()
    def labelgroup(self, prompt: str = "label", widget: str = "prompt") -> None:
        """Launch prompt widget to label the current group

        Parameters
        ==========
        prompt :
            Text with which to prompt user (default: "label")
        widget :
            Name of the prompt widget (default: "prompt")
        """

        def f(name: str) -> None:
            self.current_group.set_label(name or None)

        try:
            mb = self.widgets_map[widget]
            mb.start_input(prompt, f, allow_empty_input=True)
        except KeyError:
            logger.error("No widget named '%s' present.", widget)

    @expose_command()
    def spawncmd(
        self,
        prompt: str = "spawn",
        widget: str = "prompt",
        command: str = "%s",
        complete: str = "cmd",
        shell: bool = True,
        aliases: dict[str, str] | None = None,
    ) -> None:
        """Spawn a command using a prompt widget, with tab-completion.

        Parameters
        ==========
        prompt :
            Text with which to prompt user (default: "spawn: ").
        widget :
            Name of the prompt widget (default: "prompt").
        command :
            command template (default: "%s").
        complete :
            Tab completion function (default: "cmd")
        shell :
            Execute the command with /bin/sh (default: True)
        aliases :
            Dictionary mapping aliases to commands. If the entered command is a key in
            this dict, the command it maps to will be executed instead.
        """

        def f(args: str) -> None:
            if args:
                if aliases and args in aliases:
                    args = aliases[args]
                self.spawn(command % args, shell=shell)

        try:
            mb = self.widgets_map[widget]
            mb.start_input(prompt, f, complete, aliases=aliases)
        except KeyError:
            logger.error("No widget named '%s' present.", widget)

    @expose_command()
    def qtilecmd(
        self,
        prompt: str = "command",
        widget: str = "prompt",
        messenger: str = "xmessage",
    ) -> None:
        """Execute a Qtile command using the client syntax

        Tab completion aids navigation of the command tree

        Parameters
        ==========
        prompt :
            Text to display at the prompt (default: "command: ")
        widget :
            Name of the prompt widget (default: "prompt")
        messenger :
            Command to display output, set this to None to disable (default:
            "xmessage")
        """

        def f(cmd: str) -> None:
            if cmd:
                # c here is used in eval() below
                q = QtileCommandInterface(self)
                c = InteractiveCommandClient(q)  # noqa: F841
                try:
                    cmd_arg = str(cmd).split(" ")
                except AttributeError:
                    return
                cmd_len = len(cmd_arg)
                if cmd_len == 0:
                    logger.debug("No command entered.")
                    return
                try:
                    result = eval(f"c.{cmd:s}")
                except (CommandError, CommandException, AttributeError):
                    logger.exception("Command errored:")
                    result = None
                if result is not None:
                    from pprint import pformat

                    message = pformat(result)
                    if messenger:
                        self.spawn(f'{messenger:s} "{message:s}"')
                    logger.debug(result)

        mb = self.widgets_map[widget]
        if not mb:
            logger.error("No widget named %s present.", widget)
            return
        mb.start_input(prompt, f, "qshell")

    @expose_command()
    def addgroup(
        self,
        group: str,
        label: str | None = None,
        layout: str | None = None,
        layouts: list[Layout] | None = None,
        index: int | None = None,
        persist: bool | None = False,
    ) -> bool:
        """Add a group with the given name"""
        return self.add_group(
            name=group, layout=layout, layouts=layouts, label=label, index=index, persist=persist
        )

    @expose_command()
    def delgroup(self, group: str) -> None:
        """Delete a group with the given name"""
        self.delete_group(group)

    @expose_command()
    def add_rule(
        self,
        match_args: dict[str, Any],
        rule_args: dict[str, Any],
        min_priorty: bool = False,
    ) -> int | None:
        """Add a dgroup rule, returns rule_id needed to remove it

        Parameters
        ==========
        match_args :
            config.Match arguments
        rule_args :
            config.Rule arguments
        min_priorty :
            If the rule is added with minimum priority (last) (default: False)
        """
        if not self.dgroups:
            logger.warning("No dgroups created")
            return None

        match = Match(**match_args)
        rule = Rule([match], **rule_args)
        return self.dgroups.add_rule(rule, min_priorty)

    @expose_command()
    def remove_rule(self, rule_id: int) -> None:
        """Remove a dgroup rule by rule_id"""
        self.dgroups.remove_rule(rule_id)

    @expose_command()
    def hide_show_bar(
        self,
        position: Literal["top", "bottom", "left", "right", "all"] = "all",
        screen: Literal["current", "all"] = "current",
    ) -> None:
        """Toggle visibility of a given bar

        Parameters
        ==========
        position :
            one of: "top", "bottom", "left", "right", or "all" (default: "all")
        screen :
            one of: "current", "all" (default: "current")
        """
        to_mod = [self.current_screen]
        if screen == "all":
            to_mod = self.screens
        for s in to_mod:
            self.hide_show_bar_screen(s, position)

    def hide_show_bar_screen(
        self,
        screen: Screen,
        position: Literal["top", "bottom", "left", "right", "all"] = "all",
    ) -> None:
        if position in ["top", "bottom", "left", "right"]:
            bar = getattr(screen, position)
            if bar:
                bar.show(not bar.is_show())
                self.current_group.layout_all()
            else:
                logger.warning("Not found bar in position '%s' for hide/show.", position)
        elif position == "all":
            is_show = None
            for bar in [screen.left, screen.right, screen.top, screen.bottom]:
                if isinstance(bar, libqtile.bar.Bar):
                    if is_show is None:
                        is_show = not bar.is_show()
                    bar.show(is_show)
            if is_show is not None:
                self.current_group.layout_all()
            else:
                logger.warning("Not found bar for hide/show.")
        else:
            logger.warning("Invalid position value:%s", position)

    @expose_command()
    def get_state(self) -> str:
        """Get pickled state for restarting qtile"""
        buf = io.BytesIO()
        self.dump_state(buf)
        state = buf.getvalue().decode(errors="backslashreplace")
        logger.debug("State = %s", state)
        return state

    @expose_command()
    def tracemalloc_toggle(self) -> None:
        """Toggle tracemalloc status

        Running tracemalloc is required for `qtile top`
        """
        import tracemalloc

        if not tracemalloc.is_tracing():
            tracemalloc.start()
        else:
            tracemalloc.stop()

    @expose_command()
    def tracemalloc_dump(self) -> tuple[bool, str]:
        """Dump tracemalloc snapshot"""
        import tracemalloc

        if not tracemalloc.is_tracing():
            return False, "Trace not started"
        cache_directory = get_cache_dir()
        malloc_dump = os.path.join(cache_directory, "qtile_tracemalloc.dump")
        tracemalloc.take_snapshot().dump(malloc_dump)
        return True, malloc_dump

    @expose_command()
    def get_test_data(self) -> Any:
        """
        Returns any content arbitrarily set in the self.test_data attribute.
        Useful in tests.
        """
        return self.test_data

    @expose_command()
    def run_extension(self, extension: _Extension) -> None:
        """Run extensions"""
        extension.run()

    @expose_command()
    def fire_user_hook(self, hook_name: str, *args: Any) -> None:
        """Fire a custom hook."""
        hook.fire(f"user_{hook_name}", *args)

    @expose_command()
    def start_repl_server(self, locals_dict: dict[str, Any] = dict()) -> None:
        """Start the REPL server."""
        _locals = {"qtile": self, **locals_dict}
        create_task(repl_server.start(locals_dict=_locals))

    @expose_command()
    def stop_repl_server(self) -> None:
        """Stop the REPL server."""
        create_task(repl_server.stop())



from __future__ import annotations

import re
from abc import ABCMeta, abstractmethod
from pathlib import Path
from subprocess import CalledProcessError, check_output
from typing import TYPE_CHECKING

from libqtile.command.base import expose_command
from libqtile.confreader import ConfigError
from libqtile.log_utils import logger
from libqtile.widget import base

if TYPE_CHECKING:
    from libqtile.core.manager import Qtile


class _BaseLayoutBackend(metaclass=ABCMeta):
    def __init__(self, qtile: Qtile):
        """
        This handles getting and setter the keyboard layout with the appropriate
        backend.
        """

    @abstractmethod
    def get_keyboard(self) -> str:
        """
        Return the currently used keyboard layout as a string

        Examples: "us", "us dvorak".  In case of error returns "unknown".
        """

    def set_keyboard(self, layout: str, options: str | None) -> None:
        """
        Set the keyboard layout with specified options.
        """


class _X11LayoutBackend(_BaseLayoutBackend):
    kb_layout_regex = re.compile(r"layout:\s+(?P<layout>[\w-]+)")
    kb_variant_regex = re.compile(r"variant:\s+(?P<variant>[\w-]+)")

    def get_keyboard(self) -> str:
        try:
            command = "setxkbmap -verbose 10 -query"
            setxkbmap_output = check_output(command.split(" ")).decode()
        except CalledProcessError:
            logger.exception("Can not get the keyboard layout:")
            return "unknown"
        except OSError:
            logger.exception("Please, check that xset is available:")
            return "unknown"

        match_layout = self.kb_layout_regex.search(setxkbmap_output)
        if match_layout is None:
            return "ERR"
        keyboard = match_layout.group("layout")

        match_variant = self.kb_variant_regex.search(setxkbmap_output)
        if match_variant:
            keyboard += " " + match_variant.group("variant")
        return keyboard

    def set_keyboard(self, layout: str, options: str | None) -> None:
        command = ["setxkbmap"]
        command.extend(layout.split(" "))
        if options:
            command.extend(["-option", options])
        try:
            check_output(command)
        except CalledProcessError:
            logger.error("Cannot change the keyboard layout.")
        except OSError:
            logger.error("Please, check that setxkbmap is available.")
        else:
            # Load Xmodmap if it's available
            if Path("~/.Xmodmap").expanduser().is_file():
                try:
                    check_output("xmodmap $HOME/.Xmodmap", shell=True)
                except CalledProcessError:
                    logger.error("Could not load ~/.Xmodmap.")


class _WaylandLayoutBackend(_BaseLayoutBackend):
    def __init__(self, qtile: Qtile) -> None:
        from libqtile.backend.wayland import core as waylandcore

        assert isinstance(qtile.core, waylandcore.Core)
        self.set_keymap = qtile.core.set_keymap
        self._layout: str = ""

    def get_keyboard(self) -> str:
        return self._layout

    def set_keyboard(self, layout: str, options: str | None) -> None:
        maybe_variant: str | None = None
        if " " in layout:
            layout_name, maybe_variant = layout.split(" ", maxsplit=1)
        else:
            layout_name = layout
        self.set_keymap(layout_name, options, maybe_variant)
        self._layout = layout


layout_backends = {
    "x11": _X11LayoutBackend,
    "wayland": _WaylandLayoutBackend,
}


class KeyboardLayout(base.InLoopPollText):
    """Widget for changing and displaying the current keyboard layout

    To use this widget effectively you need to specify keyboard layouts you want to use
    (using "configured_keyboards") and bind function "next_keyboard" to specific keys in
    order to change layouts.

    For example:

        Key([mod], "space", lazy.widget["keyboardlayout"].next_keyboard(), desc="Next keyboard layout."),

    When running Qtile with the X11 backend, this widget requires setxkbmap to be available.
    Xmodmap will also be used if .Xmodmap file is available.
    """

    defaults = [
        ("update_interval", 1, "Update time in seconds."),
        (
            "configured_keyboards",
            ["us"],
            "A list of predefined keyboard layouts "
            "represented as strings. For example: "
            "['us', 'us colemak', 'es', 'fr'].",
        ),
        (
            "display_map",
            {},
            "Custom display of layout. Key should be in format "
            "'layout variant'. For example: "
            "{'us': 'us', 'lt sgs': 'sgs', 'ru phonetic': 'ru'}",
        ),
        ("option", None, "string of setxkbmap option. Ex., 'compose:menu,grp_led:scroll'"),
    ]

    def __init__(self, **config):
        base.InLoopPollText.__init__(self, **config)
        self.add_defaults(KeyboardLayout.defaults)
        self.add_callbacks({"Button1": self.next_keyboard})

    def _configure(self, qtile, bar):
        base.InLoopPollText._configure(self, qtile, bar)

        if qtile.core.name not in layout_backends:
            raise ConfigError("KeyboardLayout does not support backend: " + qtile.core.name)

        self.backend = layout_backends[qtile.core.name](qtile)
        self.backend.set_keyboard(self.configured_keyboards[0], self.option)

    @expose_command()
    def next_keyboard(self):
        """set the next layout in the list of configured keyboard layouts as
        new current layout in use

        If the current keyboard layout is not in the list, it will set as new
        layout the first one in the list.
        """

        current_keyboard = self.backend.get_keyboard()
        if current_keyboard in self.configured_keyboards:
            # iterate the list circularly
            next_keyboard = self.configured_keyboards[
                (self.configured_keyboards.index(current_keyboard) + 1)
                % len(self.configured_keyboards)
            ]
        else:
            next_keyboard = self.configured_keyboards[0]

        self.backend.set_keyboard(next_keyboard, self.option)

        self.tick()

    def poll(self):
        keyboard = self.backend.get_keyboard()
        if keyboard in self.display_map.keys():
            return self.display_map[keyboard]
        return keyboard.upper()



import libcst as cst
import libcst.matchers as m

from libqtile.scripts.migrations._base import (
    EQUALS_NO_SPACE,
    Check,
    RenamerTransformer,
    _QtileMigrator,
    add_migration,
)


class RenameWidgetTransformer(RenamerTransformer):
    from_to = ("CurrentLayoutIcon", "CurrentLayout")

    @m.leave(m.Call(func=m.Attribute(attr=m.Name("CurrentLayoutIcon"))))
    def add_kwarg_mode(self, original_node, updated_node) -> cst.Call:
        """Adds 'mode' keyword argument to 'CurrentLayout'."""
        self.lint(
            original_node,
            "CurrentLayout should add 'mode' keyword argument to have "
            "the same functionality as previous CurrentLayoutIcon.",
        )
        draw_kwarg = (
            cst.Arg(
                keyword=cst.Name("mode"),
                value=cst.SimpleString('"icon"'),
                equal=EQUALS_NO_SPACE,
            ),
        )
        new_args = draw_kwarg + original_node.args
        return updated_node.with_changes(args=new_args)


class RenameCurrentLayoutIcon(_QtileMigrator):
    ID = "RenameCurrentLayoutIcon"

    SUMMARY = "Removed ``CurrentLayoutIcon`` widget."

    HELP = """
    The ``CurrentLayoutIcon`` widget's functionality has
    been merged with ``CurrentLayout``.

    Example:

    .. code:: python

      widgets=[
          widget.CurrentLayoutIcon(),
          ...
      ],

    Should be updated as follows:

    .. code:: python

      widgets=[
          widget.CurrentLayout(mode="icon"),
          ...
      ],

    """

    AFTER_VERSION = "0.32.0"

    TESTS = [
        Check(
            """
            from libqtile import widget

            widget.CurrentLayoutIcon(font="sans")
            widget.CurrentLayoutIcon()
            widget.CurrentLayout(font="sans")
            widget.CurrentLayout()
            """,
            """
            from libqtile import widget

            widget.CurrentLayout(mode="icon", font="sans")
            widget.CurrentLayout(mode="icon")
            widget.CurrentLayout(font="sans")
            widget.CurrentLayout()
            """,
        )
    ]

    visitor = RenameWidgetTransformer


add_migration(RenameCurrentLayoutIcon)



"""
A simple IPC mechanism for communicating between two local processes. We
use marshal to serialize data - this means that both client and server must
run the same Python version, and that clients must be trusted (as
un-marshalling untrusted data can result in arbitrary code execution).
"""

from __future__ import annotations

import asyncio
import fcntl
import json
import marshal
import os.path
import socket
import struct
from typing import Any

from libqtile import hook
from libqtile.log_utils import logger
from libqtile.utils import get_cache_dir

HDRFORMAT = "!L"
HDRLEN = struct.calcsize(HDRFORMAT)

SOCKBASE = "qtilesocket.%s"


class IPCError(Exception):
    pass


def find_sockfile(display: str | None = None):
    """
    Finds the appropriate socket file for the given display.

    If unspecified, the socket file is determined as follows:

        - If WAYLAND_DISPLAY is set, use it.
        - else if DISPLAY is set, use that.
        - else check for the existence of a socket file for WAYLAND_DISPLAY=wayland-0
          and if it exists, use it.
        - else check for the existence of a socket file for DISPLAY=:0
          and if it exists, use it.
        - else raise an IPCError.

    """
    cache_directory = get_cache_dir()

    if display:
        return os.path.join(cache_directory, SOCKBASE % display)

    display = os.environ.get("WAYLAND_DISPLAY")
    if display:
        return os.path.join(cache_directory, SOCKBASE % display)

    display = os.environ.get("DISPLAY")
    if display:
        return os.path.join(cache_directory, SOCKBASE % display)

    sockfile = os.path.join(cache_directory, SOCKBASE % "wayland-0")
    if os.path.exists(sockfile):
        return sockfile

    sockfile = os.path.join(cache_directory, SOCKBASE % ":0")
    if os.path.exists(sockfile):
        return sockfile

    raise IPCError("Could not find socket file.")


class _IPC:
    """A helper class to handle properly packing and unpacking messages"""

    @staticmethod
    def unpack(data: bytes, *, is_json: bool | None = None) -> tuple[Any, bool]:
        """Unpack the incoming message

        Parameters
        ----------
        data: bytes
            The incoming message to unpack
        is_json: bool | None
            If the message should be unpacked as json.  By default, try to
            unpack json and fallback gracefully to marshalled bytes.

        Returns
        -------
        tuple[Any, bool]
            A tuple of the unpacked object and a boolean denoting if the
            message was deserialized using json.  If True, the return message
            should be packed as json.
        """
        if is_json is None or is_json:
            try:
                return json.loads(data.decode()), True
            except ValueError as e:
                if is_json:
                    raise IPCError("Unable to decode json data") from e

        try:
            assert len(data) >= HDRLEN
            size = struct.unpack(HDRFORMAT, data[:HDRLEN])[0]
            assert size >= len(data[HDRLEN:])
            return marshal.loads(data[HDRLEN : HDRLEN + size]), False
        except AssertionError as e:
            raise IPCError("error reading reply! (probably the socket was disconnected)") from e

    @staticmethod
    def pack(msg: Any, *, is_json: bool = False) -> bytes:
        """Pack the object into a message to pass"""
        if is_json:
            json_obj = json.dumps(msg, default=_IPC._json_encoder)
            return json_obj.encode()

        msg_bytes = marshal.dumps(msg)
        size = struct.pack(HDRFORMAT, len(msg_bytes))
        return size + msg_bytes

    @staticmethod
    def _json_encoder(field: Any) -> Any:
        """Convert non-serializable types to ones understood by stdlib json module"""
        if isinstance(field, set):
            return list(field)
        raise ValueError(f"Tried to JSON serialize unsupported type {type(field)}: {field}")


class Client:
    def __init__(self, socket_path: str, is_json=False) -> None:
        """Create a new IPC client

        Parameters
        ----------
        socket_path: str
            The file path to the file that is used to open the connection to
            the running IPC server.
        is_json: bool
            Pack and unpack messages as json
        """
        self.socket_path = socket_path
        self.is_json = is_json

    def call(self, data: Any) -> Any:
        return self.send(data)

    def send(self, msg: Any) -> Any:
        """Send the message and return the response from the server

        If any exception is raised by the server, that will propogate out of
        this call.
        """
        return asyncio.run(self.async_send(msg))

    async def async_send(self, msg: Any) -> Any:
        """Send the message to the server

        Connect to the server, then pack and send the message to the server,
        then wait for and return the response from the server.
        """
        try:
            reader, writer = await asyncio.wait_for(
                asyncio.open_unix_connection(path=self.socket_path), timeout=3
            )
        except (ConnectionRefusedError, FileNotFoundError):
            raise IPCError(f"Could not open {self.socket_path}")

        try:
            send_data = _IPC.pack(msg, is_json=self.is_json)
            writer.write(send_data)
            writer.write_eof()

            read_data = await asyncio.wait_for(reader.read(), timeout=10)
        except TimeoutError:
            raise IPCError("Server not responding")
        finally:
            # see the note in Server._server_callback()
            writer.close()
            await writer.wait_closed()

        data, _ = _IPC.unpack(read_data, is_json=self.is_json)

        return data


class Server:
    def __init__(self, socket_path: str, handler) -> None:
        self.socket_path = socket_path
        self.handler = handler
        self.server = None  # type: asyncio.AbstractServer | None

        # Use a flag to indicate if session is locked
        self.locked = asyncio.Event()
        hook.subscribe.locked(self.lock)
        hook.subscribe.unlocked(self.unlock)

        if os.path.exists(socket_path):
            os.unlink(socket_path)

        self.sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM, 0)
        flags = fcntl.fcntl(self.sock.fileno(), fcntl.F_GETFD)
        fcntl.fcntl(self.sock.fileno(), fcntl.F_SETFD, flags | fcntl.FD_CLOEXEC)
        self.sock.bind(self.socket_path)

    def lock(self):
        self.locked.set()

    def unlock(self):
        self.locked.clear()

    async def _server_callback(
        self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter
    ) -> None:
        """Callback when a connection is made to the server

        Read the data sent from the client, execute the requested command, and
        send the reply back to the client.
        """
        try:
            logger.debug("Connection made to server")
            data = await reader.read()
            logger.debug("EOF received by server")

            req, is_json = _IPC.unpack(data)
        except IPCError:
            logger.warning("Invalid data received, closing connection")
        else:
            # Don't handle requests when session is locked
            if self.locked.is_set():
                rep = (1, {"error": "Session locked."})
            else:
                rep = self.handler(req)

            result = _IPC.pack(rep, is_json=is_json)

            logger.debug("Sending result on receive EOF")
            writer.write(result)
            logger.debug("Closing connection on receive EOF")
            writer.write_eof()
        finally:
            writer.close()
            await writer.wait_closed()

    async def __aenter__(self) -> Server:
        """Start and return the server"""
        await self.start()
        return self

    async def __aexit__(self, _exc_type, _exc_value, _tb) -> None:
        """Close and shutdown the server"""
        await self.close()

    async def start(self) -> None:
        """Start the server"""
        assert self.server is None

        logger.debug("Starting server")
        server_coroutine = asyncio.start_unix_server(self._server_callback, sock=self.sock)
        self.server = await server_coroutine

    async def close(self) -> None:
        """Close and shutdown the server"""
        assert self.server is not None

        logger.debug("Stopping server on close")
        self.server.close()
        await self.server.wait_closed()

        self.server = None



import asyncio

import pulsectl_asyncio
from pulsectl import PulseError

from libqtile import qtile
from libqtile.command.base import expose_command
from libqtile.log_utils import logger
from libqtile.utils import create_task
from libqtile.widget.volume import VolumeBase

lock = asyncio.Lock()


class PulseConnection:
    """
    Class object to manage the connection to the pulse server and send
    volume/mute status to subscribed clients.
    """

    def __init__(self):
        self._subscribed = False
        self._event_handler = None
        self.default_sink = None
        self.default_sink_name = None
        self.pulse = None
        self.configured = False
        self.callbacks = set()
        self.qtile = qtile
        self.timer = None

    async def _configure(self):
        # Use a lock to prevent multiple connection attempts
        async with lock:
            if self.configured:
                return

            # Create pulse async object but don't connect
            self.pulse = pulsectl_asyncio.PulseAsync("qtile-pulse")

            # Try to connect
            await self._check_pulse_connection()

            self.configured = True

    async def _check_pulse_connection(self):
        """
        The PulseAsync object subscribes to connection state events so we
        need to check periodically whether the connection has been lost.
        """
        if not self.pulse.connected:
            # Check if we were previously connected to the server and,
            # if so, stop the event handler
            if self._subscribed:
                if self._event_handler is not None:
                    self._event_handler.cancel()
                    self._event_handler = None
                self._subscribed = False
            try:
                await self.pulse.connect()
                logger.debug("Connection to pulseaudio ready")
            except PulseError:
                logger.warning("Failed to connect to pulseaudio, retrying in 10s")
            else:
                # We're connected so get details of the default sink
                await self.get_server_info()

                # Start event listeners for sink and server events
                self._event_handler = create_task(self._event_listener())
                self._subscribed = True

        # Set a timer to check status in 10 seconds time
        self.timer = self.qtile.call_later(10, create_task, self._check_pulse_connection())

    async def _event_listener(self):
        """Listens for sink and server events from the server."""
        async for event in self.pulse.subscribe_events("sink", "server"):
            # Sink events will signify volume changes
            if event.facility == "sink":
                # There's been a change to available sinks
                # Update default sink details before querying the sink
                if event.t in ("new", "remove"):
                    await self.get_server_info()
                await self.get_sink_info()
            # Server events include when the default sink changes
            elif event.facility == "server":
                await self.get_server_info()

    async def get_server_info(self):
        """Updates the default sink name."""
        info = await self.pulse.server_info()
        self.default_sink_name = info.default_sink_name
        await self.get_sink_info()

    async def get_sink_info(self):
        """Gets a reference to the default sink and triggers update for subscribed clients."""
        sinks = [
            sink for sink in await self.pulse.sink_list() if sink.name == self.default_sink_name
        ]
        if not sinks:
            logger.warning("Could not get info for default sink")
            self.default_sink = None
            return

        self.default_sink = sinks[0]
        self.update_clients()

    def get_volume(self):
        """Gets volume and mute status for default sink."""
        if not self.pulse.connected:
            return None, None

        if self.default_sink:
            mute = self.default_sink.mute
            base = self.default_sink.base_volume
            if not base:
                return -1, mute
            current = self.default_sink.volume.value_flat
            return round(current * 100 / base), mute
        return -1, 0

    def update_clients(self):
        """Sends volume and mute status to subscribed clients."""
        for callback in self.callbacks:
            callback(*self.get_volume())

    def subscribe(self, callback):
        """
        Subscribes a client for callback events.

        The first subscription will trigger the connection to the
        pulse server.
        """
        need_configure = not bool(self.callbacks)
        self.callbacks.add(callback)

        if need_configure:
            create_task(self._configure())

    def unsubscribe(self, callback):
        """
        Unsubscribes a client from callback events.

        Removing the last client closes the connection with the
        pulse server and cancels future calls to connect.
        """
        self.callbacks.discard(callback)

        if not self.callbacks:
            self.pulse.close()

            # Prevent future calls to connect to the server
            if self.timer:
                self.timer.cancel()
                self.timer = None

            self.configured = False


pulse = PulseConnection()


class PulseVolume(VolumeBase):
    """
    Volume widget for systems using PulseAudio.

    The widget connects to the PulseAudio server by using the libpulse library
    and so should be updated virtually instantly rather than needing to poll the
    volume status regularly (NB this means that the ``update_interval`` parameter
    serves no purpose for this widget).

    The widget relies on the `pulsectl_asyncio <https://pypi.org/project/pulsectl-asyncio/>`__
    library to access the libpulse bindings. If you are using python 3.11 you must use
    ``pulsectl_asyncio >= 1.0.0``.

    """

    defaults = [
        ("limit_max_volume", False, "Limit maximum volume to 100%"),
    ]

    def __init__(self, **config):
        VolumeBase.__init__(self, **config)
        self.add_defaults(PulseVolume.defaults)
        self.add_callbacks(
            {
                "Button1": self.mute,
                "Button4": self.increase_vol,
                "Button5": self.decrease_vol,
            }
        )
        self.volume = 0
        self.is_mute = 0
        self._previous_state = (-1.0, -1)

    def _configure(self, qtile, bar):
        VolumeBase._configure(self, qtile, bar)
        if self.theme_path:
            self.setup_images()
        pulse.subscribe(self.get_vals)

    async def _change_volume(self, volume):
        """Sets volume on default sink."""
        await pulse.pulse.volume_set_all_chans(pulse.default_sink, volume)

    async def _mute(self):
        """Toggles mute status of default sink."""
        await pulse.pulse.sink_mute(pulse.default_sink.index, not pulse.default_sink.mute)

    @expose_command()
    def mute(self):
        """Mute the sound device."""
        create_task(self._mute())

    @expose_command()
    def increase_vol(self, value=None):
        """Increase volume."""
        if not value:
            value = pulse.default_sink.volume.value_flat + (self.step / 100.0)
        base = pulse.default_sink.base_volume
        if self.limit_max_volume and value > base:
            value = base

        create_task(self._change_volume(value))

    @expose_command()
    def decrease_vol(self, value=None):
        """Decrease volume."""
        if not value:
            value = pulse.default_sink.volume.value_flat - (self.step / 100.0)

        value = max(value, 0)

        create_task(self._change_volume(value))

    def get_vals(self, vol, muted):
        if (vol, muted) != self._previous_state:
            self.volume = vol
            self.is_mute = muted
            self._previous_state = (vol, muted)
            self.update()

    def update(self):
        """
        same method as in Volume widgets except that here we don't need to
        manually re-schedule update
        """
        if pulse.pulse is None or not pulse.pulse.connected:
            return

        # Update the underlying canvas size before actually attempting
        # to figure out how big it is and draw it.
        length = self.length
        self._update_drawer()
        if self.length == length:
            self.draw()
        else:
            self.bar.draw()

    def finalize(self):
        # Close the connection to the server
        pulse.unsubscribe(self.get_vals)
        VolumeBase.finalize(self)



from __future__ import annotations

from typing import TYPE_CHECKING

import libqtile
from libqtile.command.base import expose_command
from libqtile.layout.base import _SimpleLayoutBase

if TYPE_CHECKING:
    from libqtile.backend.base import Window
    from libqtile.config import ScreenRect


class Zoomy(_SimpleLayoutBase):
    """A layout with single active windows, and few other previews at the right"""

    defaults = [
        ("columnwidth", 150, "Width of the right column"),
        ("property_name", "ZOOM", "Property to set on zoomed window (X11 only)"),
        ("property_small", "0.1", "Property value to set on zoomed window (X11 only)"),
        ("property_big", "1.0", "Property value to set on normal window (X11 only)"),
        ("margin", 0, "Margin of the layout (int or list of ints [N E S W])"),
    ]

    def __init__(self, **config):
        _SimpleLayoutBase.__init__(self, **config)
        self.add_defaults(Zoomy.defaults)

    def add_client(self, client: Window) -> None:  # type: ignore[override]
        self.clients.append_head(client)

    def configure(self, client: Window, screen_rect: ScreenRect) -> None:
        left, right = screen_rect.hsplit(screen_rect.width - self.columnwidth)
        if client is self.clients.current_client:
            client.place(
                left.x,
                left.y,
                left.width,
                left.height,
                0,
                None,
                margin=self.margin,
            )
        else:
            h = right.width * left.height // left.width
            client_index = self.clients.index(client)
            focused_index = self.clients.current_index
            offset = client_index - focused_index - 1
            if offset < 0:
                offset += len(self.clients)
            if h * (len(self.clients) - 1) < right.height:
                client.place(
                    right.x,
                    right.y + h * offset,
                    right.width,
                    h,
                    0,
                    None,
                    margin=self.margin,
                )
            else:
                hh = (right.height - h) // (len(self.clients) - 1)
                client.place(
                    right.x,
                    right.y + hh * offset,
                    right.width,
                    h,
                    0,
                    None,
                    margin=self.margin,
                )
        client.unhide()

    def focus(self, win):
        if self.property_name and libqtile.qtile.core.name != "x11":
            self.property_name = ""

        if (
            self.clients.current_client
            and self.property_name
            and self.clients.current_client.window.get_property(self.property_name, "UTF8_STRING")
            is not None
        ):
            self.clients.current_client.window.set_property(
                self.property_name, self.property_small, "UTF8_STRING", format=8
            )
        _SimpleLayoutBase.focus(self, win)
        if self.property_name:
            win.window.set_property(
                self.property_name, self.property_big, "UTF8_STRING", format=8
            )

    @expose_command("down")
    def next(self) -> None:
        _SimpleLayoutBase.next(self)

    @expose_command("up")
    def previous(self) -> None:
        _SimpleLayoutBase.previous(self)



#!/usr/bin/env python3
"""
Command-line tool to expose qtile.command functionality to shell.
This can be used standalone or in other shell scripts.
"""

from __future__ import annotations

import argparse
import itertools
import json
import sys
import textwrap

from libqtile.command.base import CommandError, CommandException, SelectError
from libqtile.command.client import CommandClient
from libqtile.command.graph import CommandGraphRoot
from libqtile.command.interface import IPCCommandInterface
from libqtile.ipc import Client, find_sockfile


class KeyValDictAdd(argparse.Action):
    def __call__(self, parser, args, values, option_string=None):  # noqa: F841
        kwargs = {}
        for value in values:
            k, _, v = value.partition("=")
            if not (k and v):
                raise argparse.ArgumentError(self, f"{value} is not a valid key=value pair.")

            kwargs[k] = v

        setattr(args, self.dest, kwargs)


def set_to_list(obj):
    if isinstance(obj, set):
        return list(obj)
    raise TypeError


def get_formated_info(obj: CommandClient, cmd: str, args=True, short=True) -> str:
    """Get documentation for command/function and format it.

    Returns:
      * args=True, short=True - '*' if arguments are present and a summary line.
      * args=True, short=False - (function args) and a summary line.
      * args=False - a summary line.

    If 'doc' function is not present in object or there is no doc string for
    given cmd it returns empty string.  The arguments are extracted from doc[0]
    line, the summary is constructed from doc[1] line.
    """

    doc = obj.call("doc", cmd).splitlines()

    tdoc = doc[0]
    doc_args = tdoc[tdoc.find("(") : tdoc.find(")") + 1].strip()

    short_description = doc[1] if len(doc) > 1 else ""

    if not args:
        doc_args = ""
    elif short:
        doc_args = " " if doc_args == "()" else "*"

    return (doc_args + " " + short_description).rstrip()


def print_commands(prefix: str, obj: CommandClient) -> None:
    """Print available commands for given object."""
    prefix += " -f "

    cmds = obj.call("commands")

    output = []
    for cmd in cmds:
        doc_args = get_formated_info(obj, cmd)

        pcmd = prefix + cmd
        output.append([pcmd, doc_args])

    max_cmd = max(len(pcmd) for pcmd, _ in output)

    # Print formatted output
    formatting = f"{{:<{(max_cmd + 1):d}}}\t{{}}"
    for line in output:
        print(formatting.format(line[0], line[1]))


def get_object(client: CommandClient, argv: list[str]) -> CommandClient:
    """
    Constructs a path to object and returns given object (if it exists).
    """
    if argv[0] in ("cmd", "root"):
        argv = argv[1:]

    # flag noting if we have consumed arg1 as the selector, eg screen[0]
    parsed_next = False

    for arg0, arg1 in itertools.zip_longest(argv, argv[1:]):
        # previous argument was an item, skip here
        if parsed_next:
            parsed_next = False
            continue

        # check if it is an item
        try:
            client = client.navigate(arg0, arg1)
            parsed_next = True
            continue
        except SelectError:
            pass

        # check if it is an attr
        try:
            client = client.navigate(arg0, None)
            continue
        except SelectError:
            pass

        print("Specified object does not exist: " + " ".join(argv))
        sys.exit(1)

    return client


def run_function(
    client: CommandClient, funcname: str, args: list[str], kwargs: dict[str, str]
) -> str:
    "Run command with specified args on given object."
    try:
        ret = client.call(funcname, *args, **kwargs, lifted=True)
    except SelectError:
        print("error: Sorry no function ", funcname)
        sys.exit(1)
    except CommandError as e:
        print(f"error: Command '{funcname}' returned error: {str(e)}")
        sys.exit(1)
    except CommandException as e:
        print(f"error: Sorry cannot run function '{funcname}' with arguments {args}: {str(e)}")
        sys.exit(1)

    return ret


def print_base_objects() -> None:
    """Prints access objects of Client, use cmd for commands."""
    root = CommandGraphRoot()
    actions = ["-o cmd"] + [f"-o {key}" for key in root.children]
    print("Specify an object on which to execute command")
    print("\n".join(actions))


def cmd_obj(args) -> None:
    "Runs tool according to specified arguments."

    if args.obj_spec:
        sock_file = args.socket or find_sockfile()
        ipc_client = Client(sock_file)
        cmd_object = IPCCommandInterface(ipc_client)
        cmd_client = CommandClient(cmd_object)
        obj = get_object(cmd_client, args.obj_spec)

        if args.function == "help":
            try:
                print_commands("-o " + " ".join(args.obj_spec), obj)
            except CommandError:
                if len(args.obj_spec) == 1:
                    print(
                        f"{args.obj_spec} object needs a specified identifier e.g. '-o bar top'."
                    )
                    sys.exit(1)
                else:
                    raise
        elif args.info:
            print(args.function + get_formated_info(obj, args.function, args=True, short=False))
        else:
            ret = run_function(obj, args.function, args.args, args.kwargs)
            if ret is not None:
                print(json.dumps(ret, indent=2, default=set_to_list))
    else:
        print_base_objects()
        sys.exit(1)


def add_subcommand(subparsers, parents):
    epilog = textwrap.dedent(
        """\
    Examples:
     qtile cmd-obj
     qtile cmd-obj -o root # same as above
     qtile cmd-obj -o root -f prev_layout -a 3 # prev_layout on group 3
     qtile cmd-obj -o group 3 -f focus_back
     qtile cmd-obj -o root -f restart # restart qtile
    The graph traversal recurses:
     qtile cmd-obj -o screen 0 bar bottom screen group window -f info
     """
    )
    description = "Access the command interface from a shell."
    parser = subparsers.add_parser(
        "cmd-obj",
        help=description,
        parents=parents,
        epilog=epilog,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument(
        "--object",
        "-o",
        dest="obj_spec",
        nargs="+",
        default=["root"],
        help="Specify path to object (space separated).  "
        "If no --function flag display available commands.  "
        "The root node is selected by default or you can pass `root` explicitly.",
    )
    parser.add_argument("--function", "-f", default="help", help="Select function to execute.")
    parser.add_argument(
        "--args", "-a", nargs="+", default=[], help="Set arguments supplied to function."
    )
    parser.add_argument(
        "--kwargs",
        "-k",
        metavar="KEY=VALUE",
        nargs="+",
        default={},
        action=KeyValDictAdd,
        help="Set keyword arguments as key=value pairs.",
    )
    parser.add_argument(
        "--info",
        "-i",
        action="store_true",
        help="With both --object and --function args prints documentation for function.",
    )
    parser.add_argument("--socket", "-s", help="Use specified socket for IPC.")
    parser.set_defaults(func=cmd_obj)



import asyncio
import builtins
import codeop
import contextlib
import io
import re
import traceback

from libqtile.log_utils import logger
from libqtile.utils import create_task

ATTR_MATCH = re.compile(r"([\w\.]+?)(?:\.([\w]*))?$")
TERMINATOR = "___END___"
COMPLETION_REQUEST = "___COMPLETE___::"
REPL_PORT = 41414


def mark_unavailable(func):
    def _wrapper(*args, **kwargs):
        print(f"'{func.__name__}' is disabled in this REPL.")

    return _wrapper


def make_safer_env():
    """
    Returns a dict to be passed to the REPL's global environment.

    Can be used to block harmful commands.
    """

    # Interactive help blocks REPL and will cause qtile to hand
    original_help = builtins.help

    def safe_help(*args):
        """Print help on a specified object."""
        if not args:
            print("Interactive help() is disabled in this REPL.")
        else:
            return original_help(*args)

    # Store original help so we can still call it safely
    builtins.help = safe_help

    # Mask other builtins
    builtins.input = mark_unavailable(builtins.input)

    return {"__builtins__": builtins}


def parse_completion_expr(text):
    """
    Parses an input like 'qtile.win' or 'qtil' and splits it into:
    - object_expr: what to evaluate or look up ('qtile', 'qtil')
    - attr_prefix: what to complete ('', 'win', etc.)
    """
    match = ATTR_MATCH.search(text)
    if not match:
        return None, None
    obj_expr, attr_prefix = match.groups()
    return obj_expr, attr_prefix or ""


def get_completions(text, local_vars):
    expr, attr_prefix = parse_completion_expr(text)

    # Case 1: Completing a top-level variable name
    if "." not in text:
        return [name for name in local_vars if name.startswith(expr)]

    # Case 2: Completing an attribute
    try:
        base = eval(expr, {}, local_vars)
        options = [attr for attr in dir(base) if attr.startswith(attr_prefix)]
        options = [
            f"{expr}.{attr}" + ("(" if callable(getattr(base, attr)) else "") for attr in options
        ]
        options = list(filter(None, options))
        return options
    except Exception:
        return []


class QtileREPLServer:
    """
    Provides a REPL interface to allow users to inspect qtile's internals via
    a more intuitive/familiar interface compared to `qtile shell`.
    """

    def __init__(self):
        self.buffer = ""
        self.compiler = codeop.Compile()
        self.started = False
        self.connections = set()

    def evaluate_code(self, code):
        with io.StringIO() as stdout:
            # Capture any stdout and direct to a buffer
            with contextlib.redirect_stdout(stdout), contextlib.redirect_stderr(stdout):
                try:
                    try:
                        # Try eval (for expressions)
                        expr_code = compile(code, "<stdin>", "eval")
                        result = eval(expr_code, self.locals)
                        if result is not None:
                            # We can use print here as we've redirected stdout
                            print(repr(result))
                    except SyntaxError:
                        # Fallback to exec (for statements)
                        exec(self.compiler(code), self.locals)
                except Exception:
                    traceback.print_exc()

            return stdout.getvalue()

    async def handle_client(self, reader, writer):
        """Method for sending data to REPL client."""
        q = self.locals.get("qtile", None)

        async def send(message, end=True):
            """Wrapper to send data to client."""
            suffix = TERMINATOR if end else ""
            writer.write(f"{message}{suffix}\n".encode())
            await writer.drain()

        await send("Connected to Qtile REPL\nPress Ctrl+C to exit.\n")

        # Keep track of the number of connected clients so server is not
        # stopped while there is still a client connected.
        task = asyncio.current_task()
        self.connections.add(task)

        self.compiler = codeop.CommandCompiler()

        while not reader.at_eof():
            buffer = ""
            # The client handles checking when a code block is complete and
            # terminates the code with a marker. Server therefore just reads
            # until it finds that marker.
            while True:
                line = await reader.readline()
                if not line:
                    break
                line = line.decode()

                if line.strip() == TERMINATOR:
                    break

                buffer += line

            # Handle completion requests
            if buffer.startswith(COMPLETION_REQUEST):
                prefix = buffer.split("::", 1)[1]
                matches = get_completions(prefix, self.locals)
                output = ",".join(matches) + "\n"
                await send(output)
                continue

            if not buffer.strip():
                buffer = ""
                await send("", end=False)
                continue

            # Ready to execute
            output = ""

            # Block interaction if session is locked
            if q is not None and q.locked:
                output = "Server is locked."
            else:
                # Evaluate code in a thread so blocking calls don't block the eventloop
                loop = asyncio.get_running_loop()
                output = await loop.run_in_executor(None, self.evaluate_code, buffer)

            # Send output to client
            await send(output.strip())

        # Client has disconnected. Tidy up.
        writer.close()
        self.connections.remove(task)

    async def start(self, locals_dict=dict()):
        if self.started:
            return

        self.locals = {**make_safer_env(), **locals_dict}
        self.server = await asyncio.start_server(self.handle_client, "localhost", REPL_PORT)
        logger.info("Qtile REPL server running on localhost:%d", REPL_PORT)
        self.started = True

        # serve_forever() cannot be stopped except by putting it in a task and
        # cancelling that task.
        self.repl_task = create_task(self.server.serve_forever())

        try:
            await self.repl_task
        except asyncio.CancelledError:
            logger.info("Qtile REPL server has been stopped.")

    async def stop(self):
        if not self.started:
            return

        if self.connections:
            logger.debug("Can't close with active connections")
            return

        self.server.close()
        await self.server.wait_closed()
        self.repl_task.cancel()
        self.started = False


repl_server = QtileREPLServer()



from __future__ import annotations

from typing import TYPE_CHECKING

from libqtile import configurable, pangocffi

if TYPE_CHECKING:
    from cairocffi import ImageSurface

    from libqtile.backend.base import Drawer, Internal
    from libqtile.core.manager import Qtile
    from libqtile.utils import ColorType


class Popup(configurable.Configurable):
    """
    This class can be used to create popup windows that display images and/or text.
    """

    defaults = [
        ("opacity", 1.0, "Opacity of notifications."),
        ("foreground", "#ffffff", "Colour of text."),
        ("background", "#111111", "Background colour."),
        ("border", "#111111", "Border colour."),
        ("border_width", 0, "Line width of drawn borders."),
        ("font", "sans", "Font used in notifications."),
        ("fontsize", 14, "Size of font."),
        ("fontshadow", None, "Colour for text shadows, or None for no shadows."),
        ("horizontal_padding", 0, "Padding at sides of text."),
        ("vertical_padding", 0, "Padding at top and bottom of text."),
        ("text_alignment", "left", "Text alignment: left, center or right."),
        ("wrap", True, "Whether to wrap text."),
    ]

    def __init__(
        self,
        qtile: Qtile,
        x: int = 50,
        y: int = 50,
        width: int = 256,
        height: int = 64,
        **config,
    ):
        configurable.Configurable.__init__(self, **config)
        self.add_defaults(Popup.defaults)
        self.qtile = qtile

        self.win: Internal = qtile.core.create_internal(x, y, width, height)
        self.win.opacity = self.opacity
        self.win.process_button_click = self.process_button_click
        self.win.process_window_expose = self.draw

        self.drawer: Drawer = self.win.create_drawer(width, height)
        self.clear()
        self.layout = self.drawer.textlayout(
            text="",
            colour=self.foreground,
            font_family=self.font,
            font_size=self.fontsize,
            font_shadow=self.fontshadow,
            wrap=self.wrap,
            markup=True,
        )
        self.layout.layout.set_alignment(pangocffi.ALIGNMENTS[self.text_alignment])

        if self.border_width and self.border:
            self.win.paint_borders(self.border, self.border_width)

        self.x = self.win.x
        self.y = self.win.y

    def process_button_click(self, x, y, button) -> None:
        if button == 1:
            self.hide()

    @property
    def width(self) -> int:
        return self.win.width

    @width.setter
    def width(self, value: int) -> None:
        self.win.width = value
        self.drawer.width = value

    @property
    def height(self) -> int:
        return self.win.height

    @height.setter
    def height(self, value: int) -> None:
        self.win.height = value
        self.drawer.height = value

    def set_border(self, color: ColorType) -> None:
        self.win.paint_borders(color, self.border_width)

    def clear(self) -> None:
        self.drawer.clear(self.background)

    def draw_text(self, x: int | None = None, y: int | None = None) -> None:
        self.layout.draw(
            x or self.horizontal_padding,
            y or self.vertical_padding,
        )

    def draw(self) -> None:
        self.drawer.draw()

    def place(self) -> None:
        self.win.place(
            self.x, self.y, self.width, self.height, self.border_width, self.border, above=True
        )

    def unhide(self) -> None:
        self.win.unhide()

    def draw_image(self, image: ImageSurface, x: int, y: int) -> None:
        """
        Paint an image onto the window at point x, y. The image should be a surface e.g.
        loaded from libqtile.images.Img.from_path.
        """
        self.drawer.ctx.set_source_surface(image, x, y)
        self.drawer.ctx.paint()

    def hide(self) -> None:
        self.win.hide()

    def kill(self) -> None:
        self.win.kill()
        self.layout.finalize()
        self.drawer.finalize()



import asyncio
import json
import shutil

from libqtile.command.base import expose_command
from libqtile.log_utils import logger
from libqtile.utils import create_task
from libqtile.widget.base import _TextBox


class SwayNCReader:
    """
    A class to subscribe and listen to the output of the sway notification centre client.

    Clients can subscribe to the reader and will receive a parsed JSON object whenever a new message
    is received.
    """

    def __init__(self):
        self._swaync = None
        self._finalized = False
        self._process = None
        self.callbacks = []
        self.cmd = None

    def set_path(self, path):
        if self._swaync is not None and path != self._swaync:
            logger.warning("A client is trying to set a different path to swaync. Ignoring.")
            return

        self._swaync = path
        self.cmd = f"{self._swaync} -swb"

    def subscribe(self, callback):
        needs_starting = not self.callbacks
        if callback not in self.callbacks:
            self.callbacks.append(callback)

        if needs_starting:
            create_task(self.run())

    def unsubscribe(self, callback):
        if callback in self.callbacks:
            self.callbacks.remove(callback)

        if not self.callbacks and self._process is not None:
            self.stop()

    async def run(self):
        if self.cmd is None:
            return

        self._process = await asyncio.create_subprocess_shell(
            self.cmd,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.DEVNULL,
        )

        while not self._finalized:
            out = await self._process.stdout.readline()
            # process has exited so clear text and exit loop
            if not out:
                self.update("")
                self._process = None
                break
            try:
                message = json.loads(out.decode().strip())
                self.update(message)
            except Exception:
                pass

    def stop(self):
        if self._process is None:
            return

        self._process.terminate()
        self._process = None

    def update(self, msg):
        for callback in self.callbacks:
            callback(msg)


# Create a single instance of the reader.
reader = SwayNCReader()


class SwayNC(_TextBox):
    """
    A simple widget for the Sway Notification Center.

    The widget can display the number of notifications as well as the do not
    disturb status.

    Left-clicking on the widget will toggle the panel. Right-clicking will toggle the
    do not disturb status.
    """

    supported_backends = {"wayland"}

    defaults = [
        ("swaync_client", shutil.which("swaync-client"), "Command to execute."),
        (
            "dnd_status_text",
            ("DND ", ""),
            "Text to show do-not-disturb status. Tuple of text ('on', 'off').",
        ),
        ("format", "{dnd}{num}", "Text to display."),
    ]

    def __init__(self, **config):
        _TextBox.__init__(self, "", **config)
        self.add_defaults(SwayNC.defaults)
        self.add_callbacks({"Button1": self.toggle_panel, "Button3": self.toggle_dnd})

    def _configure(self, qtile, bar):
        _TextBox._configure(self, qtile, bar)
        reader.set_path(self.swaync_client)
        reader.subscribe(self._message_handler)

    def _message_handler(self, msg):
        dnd = self.dnd_status_text[0 if "dnd" in msg.get("class", "") else 1]
        num = msg.get("text", "0")
        self.update(self.format.format(dnd=dnd, num=num))

    @expose_command
    def toggle_panel(self):
        """Show swaync client panel."""
        if self.swaync_client is not None:
            self.qtile.spawn(f"{self.swaync_client} -t -sw")

    @expose_command
    def toggle_dnd(self):
        """Toggle do not disturb status."""
        if self.swaync_client is not None:
            self.qtile.spawn(f"{self.swaync_client} -d -sw")

    def finalize(self):
        reader.unsubscribe(self._message_handler)
        _TextBox.finalize(self)



from libqtile.log_utils import logger
from libqtile.widget import base


class ThermalZone(base.InLoopPollText):
    """Thermal zone widget.

    This widget was made to read thermal zone files and transform values to
    human readable format. You can set zone parameter to any standard thermal
    zone file from /sys/class/thermal directory.
    """

    orientations = base.ORIENTATION_HORIZONTAL
    defaults = [
        ("update_interval", 2.0, "Update interval"),
        ("zone", "/sys/class/thermal/thermal_zone0/temp", "Thermal zone"),
        ("format", "{temp}°C", "Display format"),
        ("format_crit", "{temp}°C CRIT!", "Critical display format"),
        ("hidden", False, "Set True to only show if critical value reached"),
        ("fgcolor_crit", "ff0000", "Font color on critical values"),
        ("fgcolor_high", "ffaa00", "Font color on high values"),
        ("fgcolor_normal", "ffffff", "Font color on normal values"),
        ("crit", 70, "Critical temperature level"),
        ("high", 50, "High themperature level"),
    ]

    def __init__(self, **config):
        super().__init__("", **config)
        self.add_defaults(ThermalZone.defaults)

    def poll(self):
        try:
            with open(self.zone) as f:
                value = round(int(f.read().rstrip()) / 1000)
        except OSError:
            logger.exception("%s does not exist", self.zone)
            return "err!"

        variables = dict()
        variables["temp"] = str(value)
        output = self.format.format(**variables)
        if value < self.high:
            self.layout.colour = self.fgcolor_normal
        elif value < self.crit:
            self.layout.colour = self.fgcolor_high
        elif value >= self.crit:
            self.layout.colour = self.fgcolor_crit
            output = self.format_crit.format(**variables)
        if self.hidden and value < self.crit:
            output = ""
        return output



from cffi import FFI
from xcffib.ffi import ffi as xcffib_ffi

xcursors_ffi = FFI()

xcursors_ffi.include(xcffib_ffi)

xcursors_ffi.cdef(
    """
    typedef uint32_t xcb_cursor_t;
    typedef struct xcb_cursor_context_t xcb_cursor_context_t;

    int xcb_cursor_context_new(
        xcb_connection_t *conn,
        xcb_screen_t *screen,
        xcb_cursor_context_t **ctx
        );

    xcb_cursor_t xcb_cursor_load_cursor(
        xcb_cursor_context_t *ctx,
        const char *name
        );

    void xcb_cursor_context_free(xcb_cursor_context_t *ctx);
"""
)
